import random
import weave

from ... import weave_internal
from . import panel_distribution


def test_flow():
    # Simulate what js will do
    # create var nodes for input and config
    items = weave.save(
        [
            {"label": "x", "values": [random.gauss(5, 2) for i in range(5)]},
            {"label": "y", "values": [random.gauss(9, 4) for i in range(5)]},
        ]
    )
    panel = panel_distribution.Distribution(
        items, value_fn=lambda x: x["values"], label_fn=lambda x: x["label"]
    )

    # JS passes weave "functions" (nodes) in, just to ensure this is a pure weave op
    input_var = weave_internal.make_var_node(items.type, "input")
    input_node = weave_internal.make_const_node(weave.type_of(input_var), input_var)
    config_var = weave_internal.make_var_node(weave.type_of(panel.config), "config")
    config_node = weave_internal.make_const_node(weave.type_of(config_var), config_var)

    rendered = weave.use(
        panel_distribution.distribution_panel_plot_render(input_node, config_node)
    )

    rendered_config = rendered.config.series[0]
    plot_select_function_x = rendered_config.table.columnSelectFunctions[
        rendered_config.dims.x
    ]
    plot_select_function_y = rendered_config.table.columnSelectFunctions[
        rendered_config.dims.y
    ]
    plot_select_function_label = rendered_config.table.columnSelectFunctions[
        rendered_config.dims.label
    ]
    assert plot_select_function_x.type == weave.types.Number()
    assert plot_select_function_y.type.value == weave.types.Int()
    assert plot_select_function_label.type == weave.types.String()



import contextlib
import contextvars
import dataclasses
import inspect
import typing
import black


from . import graph, registry_mem, codifiable_value_mixin, storage, weave_types


# User-facing Apis
def object_to_code(obj: typing.Any) -> str:
    raw_code = object_to_code_no_format(obj)
    formatted_code = black.format_str(raw_code, mode=black.FileMode())
    return formatted_code


def load(d: dict) -> typing.Any:
    return storage.from_python(d)


# Internal Apis
def object_to_code_no_format(obj: typing.Any) -> str:
    for method in [
        _try_otc_using_codifiable_mixin,  # TODO: Refactor this in followup to use duck typing.
        _try_otc_using_primitives,
        # _try_otc_using_type,  # I am pretty sure this is fragile. Skipping for now.
        _try_otc_using_nodes,
        _try_otc_using_dataclasses,
    ]:
        try:
            code = method(obj)
            if code is not None:
                return code
        except Exception as e:
            pass
            # print(e)

    # Try to use the object's storage representation:
    return _otc_using_storage_fallback(obj)


_var_node_frame: contextvars.ContextVar[
    typing.Optional[list[str]]
] = contextvars.ContextVar("var_node_frame", default=None)


@contextlib.contextmanager
def additional_var_nodes(names: list[str]) -> typing.Generator[None, None, None]:
    current_frame = _var_node_frame.get()
    if current_frame is None:
        current_frame = []
    new_frame = current_frame + names
    token = _var_node_frame.set(new_frame)
    yield
    _var_node_frame.reset(token)


def lambda_wrapped_object_to_code_no_format(
    obj: typing.Any, new_lambda_vars: list[str]
) -> str:
    if len(new_lambda_vars) == 0:
        return object_to_code_no_format(obj)
    else:
        with additional_var_nodes(new_lambda_vars):
            return (
                f"lambda {', '.join(new_lambda_vars)}: {object_to_code_no_format(obj)}"
            )


# Private Apis
def _try_otc_using_codifiable_mixin(obj: typing.Any) -> typing.Optional[str]:
    # TODO:
    #  * Make mixin a protocol.
    #  * Do this for Plot, Group, and Table.
    if isinstance(obj, codifiable_value_mixin.CodifiableValueMixin):
        return obj.to_code()
    return None


# def load_type(type_name: str, d: dict):
#     obj_type_cls = weave_types.type_name_to_type(type_name)
#     if obj_type_cls is None:
#         raise ValueError(f"Unknown type name: {type_name}")
#     return obj_type_cls().instance_from_dict(d)


# def _try_otc_using_type(obj: typing.Any) -> typing.Optional[str]:
#     # This is unlikely to work for many objects since many types will need params.
#     obj_type = weave_types.TypeRegistry.type_of(obj)
#     if obj_type is not None:
#         # check instantiation by name
#         assert weave_types.type_name_to_type(obj_type.name)() == obj_type
#         d = obj_type.instance_to_dict(obj)
#         obj_type_name = obj_type.name
#         return f"""weave.codify.load_type({obj_type_name}, {d})"""
#     return None


def _try_otc_using_primitives(obj: typing.Any) -> typing.Optional[str]:
    if isinstance(obj, (int, float, str, bool, type(None))):
        return repr(obj)
    elif isinstance(obj, list):
        return f"""[{", ".join(object_to_code_no_format(x) for x in obj)}]"""
    elif isinstance(obj, tuple):
        return f"""({", ".join(object_to_code_no_format(x) for x in obj)},)"""
    elif isinstance(obj, dict):
        return f"""{{{", ".join(f"{object_to_code_no_format(k)}: {object_to_code_no_format(v)}" for k, v in obj.items())}}}"""
    return None


def _try_otc_using_nodes(obj: typing.Any) -> typing.Optional[str]:
    if not isinstance(obj, graph.Node):
        return None

    return _node_to_code(obj)


def _try_otc_using_dataclasses(obj: typing.Any) -> typing.Optional[str]:
    if not dataclasses.is_dataclass(obj):
        return None

    class_type = type(obj)
    if class_type.__module__.startswith("weave.decorator_type") and issubclass(
        class_type, weave_types.Type
    ):
        qualified_classpath = "weave.weave_types"
        qualified_classname = f"type_name_to_type('{class_type.name}')"
    else:
        qualified_classpath = _module_name_corrections(class_type.__module__)
        qualified_classname = class_type.__name__

    if not qualified_classpath.startswith("weave."):
        return None

    fields = dataclasses.fields(obj)
    fields_to_use = {}

    allowed_init_params = inspect.signature(class_type.__init__).parameters.keys()

    for f in fields:
        if f.init:
            field_name = f.name
            if field_name in allowed_init_params:
                obj_val = getattr(obj, field_name)
                if f.default is not dataclasses.MISSING and _equality_helper(
                    obj_val, f.default
                ):
                    continue
                if f.default_factory is not dataclasses.MISSING and _equality_helper(
                    obj_val, f.default_factory()
                ):
                    continue
                fields_to_use[field_name] = obj_val

    res = f"""{qualified_classpath}.{qualified_classname}({", ".join(f"{k}={object_to_code_no_format(v)}" for k, v in fields_to_use.items())})"""
    return res


def _otc_using_storage_fallback(obj: typing.Any) -> str:
    return f"""weave.codify.load({storage.to_python(obj)})"""


# Helpers


# Hack:
def _module_name_corrections(qualified_name: str) -> str:
    if qualified_name == "weave.ops_primitives.file_local":
        return "weave.ops"
    elif qualified_name.startswith("weave.decorator_class"):
        raise ValueError("Decorator classes are not supported.")
    elif qualified_name.startswith("weave.decorator_type"):
        raise ValueError("Decorator types are not supported.")
    return qualified_name


def _equality_helper(a: typing.Any, b: typing.Any) -> bool:
    if isinstance(a, graph.Node) and isinstance(b, graph.Node):
        if isinstance(a, graph.VoidNode) and isinstance(b, graph.VoidNode):
            return True
        # TODO: More node handing here?
        return False

    return a == b


def _type_to_code(t: weave_types.Type) -> str:
    return object_to_code_no_format(t)


def _node_to_code(node: graph.Node, wrap_const_node: bool = True) -> str:
    if isinstance(node, graph.VoidNode):
        return "weave.graph.VoidNode()"
    elif isinstance(node, graph.VarNode):
        current_frame = _var_node_frame.get()

        if current_frame is not None and node.name in current_frame:
            return node.name
        return f"weave.weave_internal.make_var_node({_type_to_code(node.type)}, '{node.name}')"
    elif isinstance(node, graph.ConstNode):
        if isinstance(node.type, weave_types.Function):
            vars = list(node.type.input_types.keys())
            return lambda_wrapped_object_to_code_no_format(node.val, vars)
        else:
            val_as_code = object_to_code_no_format(node.val)
            if wrap_const_node:
                return f"weave.weave_internal.const({val_as_code})"
            else:
                return val_as_code
    elif isinstance(node, graph.OutputNode):
        full_op_name = node.from_op.name
        prefix = ""
        inputs = list(node.from_op.inputs.values())

        if node.from_op.name == "dict":
            args = ",".join(
                [
                    key + "=" + _node_to_code(val, False)
                    for key, val in node.from_op.inputs.items()
                ]
            )
            if len(node.from_op.inputs) > 0:
                args += ","
            return f"weave.ops_primitives.dict.dict_({args})"
        elif node.from_op.name == "list":
            args = ",".join(
                [
                    key + "=" + _node_to_code(val, False)
                    for key, val in node.from_op.inputs.items()
                ]
            )
            if len(node.from_op.inputs) > 0:
                args += ","
            return f"weave.ops_primitives.list_.make_list({args})"

        is_root = len(inputs) == 0 or not isinstance(
            inputs[0], (graph.OutputNode, graph.VarNode)
        )

        if is_root:
            # In this case we need to find the qualified op name
            op_def = registry_mem.memory_registry.get_op(full_op_name)
            if op_def is None:
                raise ValueError(f"Unknown op name: {full_op_name}")
            op_name = op_def.raw_resolve_fn.__name__
            prefix = _module_name_corrections(op_def.raw_resolve_fn.__module__)
        else:
            op_name = full_op_name.split("-")[-1]
            prefix = _node_to_code(inputs[0])
            inputs = inputs[1:]

        params = ",".join([_node_to_code(n, False) for n in inputs])

        # Special syntax
        if op_name == "pick":
            return f"{prefix}[{params}]"

        if len(inputs) > 0:
            params += ","

        return f"{prefix}.{op_name}({params})"
    else:
        raise ValueError(f"Unknown node type: {node}")



from fastapi import FastAPI
from modal import Image, Secret, Stub, asgi_app
from weave.artifact_wandb import WandbArtifactRef
from weave.uris import WeaveURI
from weave.deploy.util import safe_name
import os

image = (
    Image.debian_slim()
    .apt_install("git")
    .pip_install(["$REQUIREMENTS"])
    .env(
        {
            "MODEL_REF": "$MODEL_REF",
            "PROJECT_NAME": "$PROJECT_NAME",
            "AUTH_ENTITY": "$AUTH_ENTITY",
        }
    )
)
stub = Stub("$PROJECT_NAME")
uri = WeaveURI.parse("$MODEL_REF")


@stub.function(image=image, secret=Secret.from_dotenv(__file__))
@asgi_app(label=safe_name(uri.name))
def fastapi_app() -> FastAPI:
    from weave.serve_fastapi import object_method_app
    from weave.uris import WeaveURI
    from weave import api

    uri = WeaveURI.parse(os.environ["MODEL_REF"])
    uri_ref = uri.to_ref()
    if not isinstance(uri_ref, WandbArtifactRef):
        raise ValueError(f"Expected a wandb artifact ref, got {type(uri_ref)}")
    app = object_method_app(uri_ref, auth_entity=os.environ.get("AUTH_ENTITY"))

    api.init(os.environ["PROJECT_NAME"])
    # TODO: hookup / provide more control over attributes
    # with api.attributes({"env": env}):
    return app



import dataclasses
import hashlib
import json
import copy
import typing
from typing import Sequence, Sequence

from collections.abc import Mapping
from .graph_client import GraphClient
from .op_def import OpDef
from .ref_base import Ref
from . import artifact_local
from . import storage
from . import ref_base
from . import weave_types as types
from .run import RunKey, Run
from .runs import Run as WeaveRunObj


def refs_to_str(val: typing.Any) -> typing.Any:
    if isinstance(val, ref_base.Ref):
        return str(val)
    elif isinstance(val, dict):
        return {k: refs_to_str(v) for k, v in val.items()}  # type: ignore
    elif isinstance(val, list):
        return [refs_to_str(v) for v in val]  # type: ignore
    else:
        return val


def hash_inputs(
    inputs: Mapping[str, typing.Any],
) -> str:
    hasher = hashlib.md5()
    hasher.update(json.dumps(refs_to_str(inputs)).encode())
    return hasher.hexdigest()


def make_run_id(op_name: str, inputs: dict[str, typing.Any]) -> str:
    input_hash = hash_inputs(inputs)
    hasher = hashlib.md5()
    hasher.update(op_name.encode())
    hasher.update(input_hash.encode())
    return hasher.hexdigest()


@dataclasses.dataclass
class GraphClientLocal(GraphClient[WeaveRunObj]):
    ##### Read API

    # Implement the required members from the "GraphClient" protocol class
    def runs(self) -> Sequence[Run]:
        runs = storage.objects(types.RunType())
        result: list[WeaveRunObj] = []
        for run in runs:
            result.append(typing.cast(WeaveRunObj, run.get()))
        return result

    def run(self, run_id: str) -> typing.Optional[Run]:
        raise NotImplementedError

    def find_op_run(
        self, op_name: str, inputs: dict[str, typing.Any]
    ) -> typing.Optional[Run]:
        run_id = make_run_id(op_name, inputs)
        return storage.get(f"local-artifact:///run-{run_id}:latest/obj")

    def run_children(self, run_id: str) -> Sequence[Run]:
        raise NotImplementedError

    def op_runs(self, op_def: OpDef) -> Sequence[Run]:
        runs = storage.objects(types.RunType())
        result: list[WeaveRunObj] = []
        for run_ref in runs:
            run = typing.cast(WeaveRunObj, run_ref.get())
            if run.op_name == str(op_def.location):
                result.append(run)
        return result

    def ref_input_to(self, ref: Ref) -> Sequence[Run]:
        runs = storage.objects(types.RunType())
        result: list[WeaveRunObj] = []
        for run_ref in runs:
            run = typing.cast(WeaveRunObj, run_ref.get())
            for v in run.inputs.values():
                if str(ref) == str(v):
                    result.append(run)
        return result

    def ref_value_input_to(self, ref: Ref) -> list[Run]:
        runs = storage.objects(types.RunType())
        result: list[Run] = []
        for run_ref in runs:
            run = typing.cast(WeaveRunObj, run_ref.get())
            if ref.digest in run.inputs.get("_digests", []):
                result.append(run)
        return result

    def ref_output_of(self, ref: Ref) -> typing.Optional[Run]:
        runs = storage.objects(types.RunType())
        for run_ref in runs:
            run = typing.cast(WeaveRunObj, run_ref.get())
            if str(ref) == str(run.output):
                return run
        return None

    def run_feedback(self, run_id: str) -> Sequence[dict[str, typing.Any]]:
        raise NotImplementedError

    def feedback(self, feedback_id: str) -> typing.Optional[dict[str, typing.Any]]:
        raise NotImplementedError

    # Helpers

    def ref_is_own(self, ref: typing.Optional[ref_base.Ref]) -> bool:
        return isinstance(ref, artifact_local.LocalArtifactRef)

    def ref_uri(
        self, name: str, version: str, path: str
    ) -> artifact_local.WeaveLocalArtifactURI:
        return artifact_local.WeaveLocalArtifactURI(name, version, path=path)

    def run_ui_url(self, run: Run) -> str:
        raise NotImplementedError

    ##### Write API

    def save_object(
        self, obj: typing.Any, name: str, branch_name: str
    ) -> artifact_local.LocalArtifactRef:
        from . import storage

        return storage.direct_save(
            obj,
            name=name,
            branch_name=branch_name,
        )

    def create_run(
        self,
        op_name: str,
        parent: typing.Optional["RunKey"],
        inputs: typing.Dict[str, typing.Any],
        input_refs: Sequence[Ref],
    ) -> WeaveRunObj:
        run_id = make_run_id(op_name, inputs)
        with_digest_inputs = copy.copy(inputs)
        with_digest_inputs["_digests"] = [ref.digest for ref in input_refs]
        return WeaveRunObj(run_id, op_name, inputs=with_digest_inputs)

    def fail_run(self, run: Run, exception: BaseException) -> None:
        # TODO: Need to implement for local, for now just do nothing.
        pass

    def finish_run(
        self,
        run: WeaveRunObj,
        output: typing.Any,
        output_refs: Sequence[Ref],
    ) -> None:
        run.output = self.save_object(output, f"run-{run.id}-output", "latest")
        ref_base._put_ref(output, run.output)
        self.save_object(run, f"run-{run.id}", "latest")

    def add_feedback(self, run_id: str, feedback: typing.Any) -> None:
        raise NotImplementedError



import os
from pathlib import Path
import sys
import shutil
import string
import subprocess
import typing
import tempfile

from weave.artifact_wandb import WeaveWBArtifactURI
from ..util import execute, safe_name
from weave import environment, __version__
from weave.uris import WeaveURI


def generate_dockerfile(
    model_ref: str,
    model_method: typing.Optional[str] = None,
    project_name: typing.Optional[str] = None,
    auth_entity: typing.Optional[str] = None,
    base_image: typing.Optional[str] = "python:3.11",
) -> str:
    """Generates a Dockerfile to run the weave op"""
    if project_name is None:
        ref_uri = WeaveURI.parse(model_ref)
        if not isinstance(ref_uri, WeaveWBArtifactURI):
            raise ValueError(f"Expected a wandb artifact ref, got {type(ref_uri)}")
        project_name = ref_uri.project_name
    src = Path(__file__).parent.parent / "Dockerfile"
    template = string.Template(src.read_text())

    return template.substitute(
        {
            "PROJECT_NAME": project_name,
            "BASE_IMAGE": base_image,
            "MODEL_REF": model_ref,
            "MODEL_METHOD": model_method or "",
            "AUTH_ENTITY": auth_entity or "",
        }
    )


def generate_requirements_txt(model_ref: str, dir: str, dev: bool = False) -> str:
    """Generate a requirements.txt file."""
    cwd = Path(os.getcwd())
    if dev and (cwd / "build_dist.py").exists():
        print("Building weave for development...")
        env = os.environ.copy()
        env.update({"WEAVE_SKIP_BUILD": "1"})
        execute([sys.executable, str(cwd / "build_dist.py")], env=env, capture=False)
        wheel = f"weave-{__version__}-py3-none-any.whl"
        execute(["cp", str(cwd / "dist" / wheel), dir], capture=False)
        weave = f"/app/{wheel}"
    else:
        weave = "weave @ git+https://github.com/wandb/weave@master"
    # TODO: add any additional reqs the op needs

    # We're requiring faiss-cpu for now here, to get Hooman Slackbot deploy
    # working. But this is not right, objects and ops should have their own
    # requirements that we compile together here.
    # TODO: Fix
    return f"""
uvicorn[standard]
fastapi
faiss-cpu
{weave}
"""


def gcloud(
    args: list[str],
    timeout: typing.Optional[float] = None,
    input: typing.Optional[str] = None,
    capture: bool = True,
) -> typing.Any:
    gcloud_absolute_path = shutil.which("gcloud")
    if gcloud_absolute_path is None:
        raise ValueError(
            "gcloud command required: https://cloud.google.com/sdk/docs/install"
        )
    if os.getenv("DEBUG") == "true":
        print(f"Running gcloud {' '.join(args)}")
    return execute(
        [gcloud_absolute_path] + args, timeout=timeout, capture=capture, input=input
    )


def enforce_login() -> None:
    """Ensure the user is logged in to gcloud."""
    try:
        auth = gcloud(["auth", "print-access-token", "--format=json"], timeout=3)
        if auth.get("token") is None:
            raise ValueError()
    except (subprocess.TimeoutExpired, ValueError):
        raise ValueError("Not logged in to gcloud. Please run `gcloud auth login`.")


def compile(
    model_ref: str,
    model_method: typing.Optional[str] = None,
    wandb_project: typing.Optional[str] = None,
    auth_entity: typing.Optional[str] = None,
    base_image: typing.Optional[str] = None,
    dev: bool = False,
) -> str:
    """Compile the weave application."""
    dir = tempfile.mkdtemp()
    reqs = os.path.join(dir, "requirements")
    os.mkdir(reqs)
    with open(os.path.join(reqs, "requirements.txt"), "w") as f:
        f.write(generate_requirements_txt(model_ref, reqs, dev))
    with open(os.path.join(dir, "Dockerfile"), "w") as f:
        f.write(
            generate_dockerfile(
                model_ref, model_method, wandb_project, auth_entity, base_image
            )
        )
    return dir


def ensure_service_account(
    name: str = "weave-default", project: typing.Optional[str] = None
) -> str:
    """Ensure the user has a service account."""
    if len(name) < 6 or len(name) > 30:
        raise ValueError("Service account name must be between 6 and 30 characters.")
    project = project or gcloud(["config", "get", "project", "--format=json"])
    account = gcloud(["auth", "list", "--filter=status:ACTIVE", "--format=json"])[0][
        "account"
    ]
    sa = f"{name}@{project}.iam.gserviceaccount.com"
    exists = gcloud(
        ["iam", "service-accounts", "list", f"--filter=email={sa}", "--format=json"]
    )
    if len(exists) == 0:
        print(f"Creating service account {name}...")
        display_name = (
            " ".join([n.capitalize() for n in name.split("-")]) + " Service Account"
        )
        gcloud(
            [
                "iam",
                "service-accounts",
                "create",
                name,
                f"--display-name={display_name}",
                f"--project={project}",
                "--format=json",
            ]
        )
        gcloud(
            [
                "iam",
                "service-accounts",
                "add-iam-policy-binding",
                f"{sa}",
                f"--project={project}",
                f"--member=user:{account}",
                "--role=roles/iam.serviceAccountUser",
                "--format=json",
            ]
        )
        print(
            "To grant additional permissions, run add-iam-policy-binding on the resource:"
        )
        print(
            "  gcloud storage buckets add-iam-policy-binding gs://BUCKET --member=serviceAccount:{sa} --role=ROLE"
        )
    else:
        print(f"Using service account: {sa}")
    return sa


def ensure_secret(
    name: str, value: str, service_account: str, project: typing.Optional[str] = None
) -> None:
    """Ensure a secret exists and is accessbile by the service account."""
    project = project or gcloud(["config", "get", "project", "--format=json"])
    exists = gcloud(
        [
            "secrets",
            "list",
            f"--filter=name~^.*\/{name}$",
            f"--project={project}",
            "--format=json",
        ]
    )
    if len(exists) == 0:
        print(f"Creating secret {name} and granting access to {service_account}...")
        gcloud(
            [
                "secrets",
                "create",
                name,
                f"--project={project}",
                "--replication-policy=automatic",
                "--format=json",
            ]
        )
    # To support changing service accounts, we always add secretAccessor
    gcloud(
        [
            "secrets",
            "add-iam-policy-binding",
            name,
            f"--project={project}",
            "--format=json",
            f"--member=serviceAccount:{service_account}",
            "--role=roles/secretmanager.secretAccessor",
        ]
    )
    gcloud(
        [
            "secrets",
            "versions",
            "add",
            name,
            f"--project={project}",
            "--format=json",
            "--data-file=-",
        ],
        input=value,
    )


# This is a sketch or the commands needed to downscope permissions and use secrets
def deploy(
    model_ref: str,
    model_method: typing.Optional[str] = None,
    wandb_project: typing.Optional[str] = None,
    gcp_project: typing.Optional[str] = None,
    region: typing.Optional[str] = None,
    service_account: typing.Optional[str] = None,
    auth_entity: typing.Optional[str] = None,
    base_image: typing.Optional[str] = "python:3.11",
    memory: typing.Optional[str] = "500Mi",
) -> None:
    """Deploy the weave application."""
    enforce_login()
    if region is None:
        region = gcloud(["config", "get", "compute/region", "--format=json"])
        if region is []:
            raise ValueError(
                "No default region set. Run `gcloud config set functions/region <region>` or set the region argument."
            )
    if service_account is None:
        try:
            service_account = ensure_service_account(project=gcp_project)
        except ValueError:
            print(
                "WARNING: No service account specified.  Using the compute engine default service account..."
            )
    dir = compile(model_ref, model_method, wandb_project, auth_entity, base_image)
    ref = WeaveURI.parse(model_ref)
    if not isinstance(ref, WeaveWBArtifactURI):
        raise ValueError(f"Expected a wandb artifact ref, got {type(ref)}")
    name = safe_name(f"{ref.project_name}-{ref.name}")
    project = wandb_project or ref.project_name
    key = environment.weave_wandb_api_key()
    secrets = {
        "WANDB_API_KEY": key,
    }
    if os.getenv("OPENAI_API_KEY"):
        secrets["OPENAI_API_KEY"] = os.getenv("OPENAI_API_KEY")
    args = [
        "run",
        "deploy",
        name,
        f"--region={region}",
        f"--memory={memory}",
        f"--set-env-vars=PROJECT_NAME={project}",
        f"--source={dir}",
        "--allow-unauthenticated",
    ]
    sec_or_env = ""
    if service_account is not None:
        args.append(f"--service-account={service_account}")
        sec_or_env = "--set-secrets="
        for k, v in secrets.items():
            if v is not None:
                ensure_secret(k, v, service_account, gcp_project)
                sec_or_env += f"{k}={k}:latest,"
    else:
        sec_or_env = "--set-env-vars="
        for k, v in secrets.items():
            sec_or_env += f"{k}={v},"
    # trim the trailing comma
    sec_or_env = sec_or_env[:-1]
    args.append(sec_or_env)
    if gcp_project is not None:
        args.append(f"--project={gcp_project}")
    gcloud(args, capture=False)
    shutil.rmtree(dir)


def develop(
    model_ref: str,
    model_method: typing.Optional[str] = None,
    base_image: typing.Optional[str] = "python:3.11",
    auth_entity: typing.Optional[str] = None,
) -> None:

    dir = compile(
        model_ref,
        model_method=model_method,
        base_image=base_image,
        auth_entity=auth_entity,
        dev=True,
    )
    name = safe_name(WeaveURI.parse(model_ref).name)
    docker = shutil.which("docker")
    if docker is None:
        raise ValueError("docker command required: https://docs.docker.com/get-docker/")
    print("Building container from: ", dir)
    execute(
        [docker, "buildx", "build", "-t", name, "--load", "."], cwd=dir, capture=False
    )
    env_api_key = environment.weave_wandb_api_key()
    if env_api_key is None:
        raise ValueError("WANDB_API_KEY environment variable required")
    env = {"WANDB_API_KEY": env_api_key}
    env.update(os.environ.copy())
    print("Running container at http://localhost:8080")
    execute(
        [
            docker,
            "run",
            "-p",
            "8080:8080",
            "-e",
            "WANDB_API_KEY",
            "-e",
            "OPENAI_API_KEY",
            name,
        ],
        env=env,
        capture=False,
    )
    if os.getenv("DEBUG") == None:
        print("Cleaning up...")
        shutil.rmtree(dir)



import dataclasses
import typing

import weave
from weave import weave_internal
from weave.panel_util import make_node

from . import weave_plotly
from ...language_features.tagging import tagged_value_type

TIME_SERIES_BIN_SIZES_SEC = [
    # TODO: will need more steps along here for smooth zooming.
    1e-9,  # ns
    1e-6,  # microsec
    1e-3,  # ms
    1,
    2.5,
    5,
    10,
    15,
    30,
    60,  # 1min
    300,  # 5min
    600,  # 10min
    1200,  # 20min
    1800,  # 30 min
    *(3600 * i for i in range(1, 25)),  # 1 - 24 hr, increments of 1hr
    *(86400 * i for i in range(2, 31)),  # 2 - 30 days, increments of 1 day
    *(
        86400 * 30 * i for i in range(2, 13)
    ),  # 2 - 12 months (assuming 1 month = 30days) increments of 1 month
    *(
        365 * 86400 * i for i in range(1, 11)
    ),  # 1 - 10 years (assuming 1 year = 365 days) increments of 1 year
]

TIME_SERIES_BIN_SIZES_SEC_NODE = weave_internal.make_const_node(
    weave.types.List(weave.types.Number()), TIME_SERIES_BIN_SIZES_SEC
)

N_BINS = 50  # number of bins to show in plot

mark_py_type = typing.Union[
    typing.Literal["line"], typing.Literal["bar"], typing.Literal["point"]
]

mark_weave_type = weave.types.UnionType(
    weave.types.Const(weave.types.String(), "line"),
    weave.types.Const(weave.types.String(), "bar"),
    weave.types.Const(weave.types.String(), "point"),
)


@weave.type()
class TimeSeriesConfig:
    x: weave.Node[typing.Any] = dataclasses.field(
        default_factory=lambda: weave.graph.VoidNode()
    )
    agg: weave.Node[typing.Any] = dataclasses.field(
        default_factory=lambda: weave.graph.VoidNode()
    )
    min_x: weave.Node[typing.Any] = dataclasses.field(
        default_factory=lambda: weave.graph.VoidNode()
    )
    max_x: weave.Node[typing.Any] = dataclasses.field(
        default_factory=lambda: weave.graph.VoidNode()
    )
    label: weave.Node[
        typing.Optional[typing.Union[str, weave.types.InvalidPy]]
    ] = dataclasses.field(default_factory=lambda: weave.graph.VoidNode())
    mark: weave.Node[str] = dataclasses.field(
        default_factory=lambda: weave.graph.ConstNode(weave.types.String(), "bar")
    )
    axis_labels: weave.Node[dict[str, str]] = dataclasses.field(
        default_factory=lambda: weave.graph.ConstNode(
            weave.types.Dict(weave.types.String(), weave.types.String()),
            {},
        )
    )


def first_column_of_type(
    node_type: weave.types.Type,
    desired_type: weave.types.Type,
) -> typing.Tuple[weave.graph.ConstNode, weave.graph.ConstNode]:
    if isinstance(node_type, tagged_value_type.TaggedValueType):
        node_type = node_type.value
    if weave.types.List().assign_type(node_type):
        node_type = typing.cast(weave.types.List, node_type)
        object_type = node_type.object_type
        if desired_type.assign_type(object_type):
            return weave_internal.define_fn(
                {"input_node": node_type}, lambda item: item
            ), weave_internal.define_fn({"item": object_type}, lambda item: item)
        elif weave.types.TypedDict().assign_type(object_type):
            object_type = typing.cast(weave.types.TypedDict, object_type)
            _, non_none_desired = weave.types.split_none(desired_type)
            if (
                isinstance(non_none_desired, weave.types.Timestamp)
                and "_timestamp" in object_type.property_types
            ):
                return (
                    weave_internal.define_fn(
                        {"input_node": node_type},
                        lambda item: (item["_timestamp"] * 1000).toTimestamp(),
                    ),
                    weave_internal.define_fn(
                        {"item": object_type},
                        lambda item: (item["_timestamp"] * 1000).toTimestamp(),
                    ),
                )
            for key in object_type.property_types:
                value_type = object_type.property_types[key]
                if desired_type.assign_type(value_type):
                    return weave_internal.define_fn(
                        {"input_node": node_type}, lambda item: item[key]
                    ), weave_internal.define_fn(
                        {"item": object_type}, lambda item: item[key]
                    )
        # return weave_internal.define_fn(
        #     {"input_node": node_type}, weave.graph.VoidNode()
        # ), weave_internal.define_fn({"item": object_type}, lambda _: weave.graph.VoidNode())
    raise ValueError(
        f"Can't extract column with type {desired_type} from node of type {node_type}"
    )


@weave.op(
    input_type={
        "fn": weave.types.Function(
            {
                "item": weave.types.Any(),
            },
            weave.types.Any(),
        )
    }
)
def function_to_string(fn) -> str:
    ret = str(fn)
    if fn.val.from_op.name.endswith("pick"):
        ret = str(fn.val.from_op.inputs["key"])
    elif fn.val.from_op.name.endswith("__getattr__"):
        ret = str(fn.val.from_op.inputs["attr"])
    ret = ret.strip('"')
    return ret


# The render op. This renders the panel.


# The interface for constructing this Panel from Python
@weave.type()
class TimeSeries(weave.Panel):
    id = "TimeSeries"
    input_node: weave.Node[list[typing.Any]]
    config: typing.Optional[TimeSeriesConfig] = None

    def __init__(self, input_node, vars=None, config=None, **options):
        super().__init__(input_node=input_node, vars=vars)
        self.config = config

        unnested = weave.ops.unnest(input_node)

        # TODO: add the ability to configure options here
        if self.config is None:
            self.config = TimeSeriesConfig()
            for attr in ["x", "min_x", "max_x", "label", "mark", "agg", "axis_labels"]:
                if attr in options:
                    value = options[attr]
                    if not isinstance(value, weave.graph.Node):
                        if attr in ["min_x", "max_x", "mark", "axis_labels"]:
                            value = make_node(value)
                        if attr in ["min_x", "max_x"]:
                            value = weave_internal.const(value)
                            # value = weave.make_node(value)
                        elif attr in ["x", "label"]:
                            value = weave_internal.define_fn(
                                {"item": unnested.type.object_type}, value
                            )
                        elif attr in ["agg"]:
                            value = weave_internal.define_fn(
                                {"group": weave.types.List(unnested.type.object_type)},
                                value,
                            )
                else:
                    value = weave_internal.define_fn(
                        {"item": unnested.type.object_type},
                        lambda item: weave.graph.VoidNode(),
                    )
                setattr(self.config, attr, value)

    @weave.op()
    def initialize(self) -> TimeSeriesConfig:
        input_node = self.input_node
        # TODO: unnested_table should be Node, that way we don't evaluate it just
        # to figure it outs type here.
        # TODO: we need input variables (frame) available here. For now we have
        # manually construct them :(
        col_fn, item_fn = first_column_of_type(
            input_node.type, weave.types.optional(weave.types.Timestamp())
        )

        min_x_called = col_fn(input_node).min()  # type: ignore
        min_x = weave_internal.const(
            min_x_called,
            weave.types.Function(col_fn.type.input_types, min_x_called.type),  # type: ignore
        )

        max_x_called = col_fn(input_node).max()  # type: ignore
        max_x = weave_internal.const(
            max_x_called,
            weave.types.Function(col_fn.type.input_types, max_x_called.type),  # type: ignore
        )

        mark = weave_internal.const("bar")

        config = TimeSeriesConfig(
            x=item_fn,
            label=first_column_of_type(
                input_node.type,
                weave.types.optional(
                    weave.types.union(
                        weave.types.String(),
                        weave.types.Boolean(),
                    )
                ),
            )[1],
            agg=weave_internal.define_fn(
                {"group": input_node.type},  # type: ignore
                lambda group: group.count(),
            ),
            min_x=min_x,
            max_x=max_x,
            mark=mark,
        )

        return config

    # The config render op. This renders the config editor.
    @weave.op()
    def render_config(self) -> weave.panels.Group:
        input_node = self.input_node
        config = typing.cast(TimeSeriesConfig, self.config)
        return weave.panels.Group(
            items={
                "x": weave.panels.LabeledItem(
                    label="x", item=weave.panels.FunctionEditor(config.x)
                ),
                "label": weave.panels.LabeledItem(
                    label="label", item=weave.panels.FunctionEditor(config.label)
                ),
                "min_x": weave.panels.LabeledItem(
                    label="min_x", item=weave.panels.FunctionEditor(config.min_x)
                ),
                "max_x": weave.panels.LabeledItem(
                    label="max_x", item=weave.panels.FunctionEditor(config.max_x)
                ),
                "agg": weave.panels.LabeledItem(
                    label="agg", item=weave.panels.FunctionEditor(config.agg)
                ),
                "mark": weave.panels.LabeledItem(
                    label="mark",
                    item=weave.panels.ObjectPicker(
                        weave_internal.make_const_node(
                            weave.types.List(weave.types.String()),
                            [
                                "bar",
                                "line",
                                "point",
                            ],
                        ),
                        config=weave.panels.ObjectPickerConfig(choice=config.mark),
                    ),
                ),
            }
        )

    @weave.op()
    def render(self) -> weave_plotly.PanelPlotly:
        input_node = self.input_node
        config = typing.cast(TimeSeriesConfig, self.config)
        # NOTE: everything inside this function is operating on nodes. There are no concrete values in here.
        # We are just constructing a graph here. It will be executed later on.

        unnested = input_node.unnest()  # type: ignore

        min_x = config.min_x
        max_x = config.max_x

        if not weave.types.optional(weave.types.Timestamp()).assign_type(
            min_x.type
        ) or not weave.types.optional(weave.types.Timestamp()).assign_type(max_x.type):
            return weave.panels.PanelHtml(weave.ops.Html("No data"))  # type: ignore

        exact_bin_size = ((max_x - min_x) / N_BINS).totalSeconds()  # type: ignore
        bin_size_index = TIME_SERIES_BIN_SIZES_SEC_NODE.map(  # type: ignore
            lambda x: (
                (x - exact_bin_size).abs()
                / weave.ops.make_list(a=x, b=exact_bin_size).min()
            )
            # lambda x: (x / exact_bin_size - 1).abs() # original
        ).argmin()

        bin_size = TIME_SERIES_BIN_SIZES_SEC_NODE[bin_size_index]  # type: ignore
        bin_size_s = bin_size

        def bin_fn(item):
            label_fn_output_type = config.label.type.output_type
            group_items = {}

            # convert timestamp to seconds
            timestamp = config.x(item)

            bin_start_ms = timestamp.floor(bin_size_s)
            bin_end_ms = timestamp.ceil(bin_size_s)

            bin_start = bin_start_ms
            bin_end = bin_end_ms

            group_items["bin"] = weave.ops.dict_(start=bin_start, stop=bin_end)

            if label_fn_output_type == weave.types.Invalid():
                group_items["label"] = "no_label"
            else:
                group_items["label"] = config.label(item)

            return weave.ops.dict_(**group_items)

        binned = (
            unnested.filter(
                lambda item: weave.ops.Boolean.bool_and(
                    config.x(item) <= max_x, config.x(item) >= min_x  # type: ignore
                )
            )
            .groupby(lambda item: bin_fn(item))
            .map(
                lambda group: weave.ops.dict_(
                    x=group.groupkey()["bin"],
                    label=group.groupkey()["label"],
                    y=config.agg(group),  # type: ignore
                )
            )
            # this is needed because otherwise the lines look like a scrambled mess
            .sort(lambda item: weave.ops.make_list(a=item["x"]["start"]), ["asc"])
        )

        default_labels = weave.ops.dict_(
            # x=function_to_string(config.x),
            # y=function_to_string(config.agg),
            # label=function_to_string(config.label),
            x="x",
            y="y",
            label="label",
        )

        fig = weave_plotly.plotly_time_series(
            binned, config.mark, default_labels, config.axis_labels
        )
        return weave_plotly.PanelPlotly(fig)



import typing
import dateutil
import dateutil.parser
from ..api import op, type
from .. import weave_types as types
import datetime


@op(
    name="datetime-sub",
    input_type={
        "lhs": types.Timestamp(),
        "rhs": types.optional(types.Timestamp()),
    },
    output_type=types.optional(types.TimeDelta()),
)
def datetime_sub(lhs, rhs):
    if rhs == None:
        return None
    return lhs - rhs


@op(
    name="datetimetd-sub",
    input_type={
        "lhs": types.Timestamp(),
        "rhs": types.optional(types.TimeDelta()),
    },
    output_type=types.optional(types.Timestamp()),
)
def datetimetd_sub(lhs, rhs):
    if rhs == None:
        return None
    return lhs - rhs


@op(
    name="datetime-addms",
    input_type={
        "lhs": types.UnionType(types.Timestamp(), types.Number()),
        "rhs": lambda input_types: types.optional(
            types.Timestamp()
            if types.Number().assign_type(input_types["lhs"])
            else types.Number()
        ),
    },
    output_type=types.optional(types.Timestamp()),
)
def datetime_addms(lhs, rhs):
    if rhs == None:
        return None
    return lhs + datetime.timedelta(milliseconds=rhs)


@op(
    name="datetime-add",
    input_type={
        "lhs": types.UnionType(types.Timestamp(), types.TimeDelta()),
        "rhs": lambda input_types: types.optional(
            types.Timestamp()
            if types.TimeDelta().assign_type(input_types["lhs"])
            else types.TimeDelta()
        ),
    },
    output_type=types.optional(types.Timestamp()),
)
def datetime_add(lhs, rhs):
    if rhs == None:
        return None
    return lhs + rhs


AUTO_FORMAT_UNITS_AND_NUM_MS = (
    ("years", 365 * 24 * 60 * 60 * 1000),
    ("months", 30 * 24 * 60 * 60 * 1000),  # Approximating a month as 30 days
    ("days", 24 * 60 * 60 * 1000),
    ("hours", 60 * 60 * 1000),
    ("minutes", 60 * 1000),
    ("seconds", 1000),
    ("milliseconds", 1),
)


@op(
    name="timestamp-relativeStringAutoFormat",
    input_type={
        "timestamp1": types.Timestamp(),
        "timestamp2": types.optional(types.Timestamp()),
    },
    output_type=types.optional(types.String()),
)
def auto_format_relative_string(timestamp1, timestamp2):
    if timestamp2 == None:
        return None

    delta: datetime.timedelta = timestamp1 - timestamp2
    diff_ms = delta.total_seconds() * 1000

    for unit, unit_ms in AUTO_FORMAT_UNITS_AND_NUM_MS:
        if abs(diff_ms) >= unit_ms:
            rounding_unit = 1 if unit == "years" or unit == "months" else 0
            diff = round(diff_ms / unit_ms, rounding_unit)

            if int(diff) == diff:
                diff = int(diff)

            if abs(diff) == 1:
                unit = unit[:-1]

            return f"{diff} {unit}"

    return "less than 1 ms"


@op(
    name="timedelta-mult",
    input_type={
        "lhs": types.UnionType(types.Number(), types.TimeDelta()),
        "rhs": lambda input_types: types.optional(
            types.TimeDelta()
            if types.Number().assign_type(input_types["lhs"])
            else types.Number()
        ),
    },
    output_type=types.optional(types.TimeDelta()),
)
def timedelta_mult(lhs, rhs):
    if rhs == None:
        return None
    return lhs * rhs


@op(
    name="timedelta-div",
    input_type={"lhs": types.TimeDelta(), "rhs": types.optional(types.Number())},
    output_type=types.optional(types.TimeDelta()),
)
def timedelta_div(lhs, rhs):
    if rhs == None:
        return None
    return lhs / rhs


@op(
    name="datetime-__le__",
    input_type={
        "lhs": types.optional(types.Timestamp()),
        "rhs": types.optional(types.Timestamp()),
    },
    output_type=types.Boolean(),
)
def datetime_le(lhs, rhs):
    if rhs == None:
        return None
    return lhs <= rhs


@op(
    name="datetime-__lt__",
    input_type={
        "lhs": types.optional(types.Timestamp()),
        "rhs": types.optional(types.Timestamp()),
    },
    output_type=types.Boolean(),
)
def datetime_lt(lhs, rhs):
    if rhs == None:
        return None
    return lhs < rhs


@op(
    name="datetime-__ge__",
    input_type={
        "lhs": types.optional(types.Timestamp()),
        "rhs": types.optional(types.Timestamp()),
    },
    output_type=types.Boolean(),
)
def datetime_ge(lhs, rhs):
    if rhs == None:
        return None
    return lhs >= rhs


@op(
    name="datetime-__gt__",
    input_type={
        "lhs": types.optional(types.Timestamp()),
        "rhs": types.optional(types.Timestamp()),
    },
    output_type=types.Boolean(),
)
def datetime_gt(lhs, rhs):
    if rhs == None:
        return None
    return lhs > rhs


@op(
    name="timedelta-totalSeconds",
    input_type={"td": types.TimeDelta()},
    output_type=types.Number(),
)
def timedelta_total_seconds(td):
    return td.total_seconds()


@op(
    name="date-toNumber",
    input_type={"date": types.union(types.Timestamp(), types.LegacyDate())},
    output_type=types.Number(),
)
def to_number(date):
    return int(date.timestamp()) * 1000


@op(
    name="date-fromNumber",
    input_type={"number": types.Number()},
    output_type=types.Timestamp(),
)
def from_number(number):
    return datetime.datetime.fromtimestamp(number)


@op(
    name="date-floor",
    input_type={"date": types.union(types.Timestamp(), types.LegacyDate())},
    output_type=types.Timestamp(),
)
def floor(date, multiple_s: float):
    seconds = (date.replace(tzinfo=None) - date.min).seconds
    rounding = (seconds // multiple_s) * multiple_s
    return date + datetime.timedelta(0, rounding - seconds, -date.microsecond)


@op(
    name="date-ceil",
    input_type={"date": types.union(types.Timestamp(), types.LegacyDate())},
    output_type=types.Timestamp(),
)
def ceil(date, multiple_s: float):
    seconds = (date.replace(tzinfo=None) - date.min).seconds
    rounding = (seconds // multiple_s) * multiple_s
    return date + datetime.timedelta(
        0, rounding - seconds + multiple_s, -date.microsecond
    )


@op(
    name="date_round-month",
    input_type={"date": types.union(types.Timestamp(), types.LegacyDate())},
    output_type=types.Timestamp(),
)
def round_month(date):
    return datetime.datetime(date.year, date.month, 1, 0, 0, 0, 0, date.tzinfo)


@op(
    name="date_round-week",
    input_type={"date": types.union(types.Timestamp(), types.LegacyDate())},
    output_type=types.Timestamp(),
)
def round_week(date):
    curr_weekday = date.weekday()  # MON = 0, SUN = 6
    prev_sunday_offset = (
        curr_weekday + 1
    ) % 7  # MON = 0, SUN = 6 -> SUN = 0 .. SAT = 6
    new_date = datetime.datetime(
        date.year, date.month, date.day, 0, 0, 0, 0, date.tzinfo
    )
    new_date -= datetime.timedelta(days=prev_sunday_offset)
    return new_date


@op(
    name="dates-equal",
    input_type={
        "lhs": types.union(types.Timestamp(), types.LegacyDate()),
        "rhs": types.union(types.Timestamp(), types.LegacyDate()),
    },
    output_type=types.Boolean(),
)
def dates_equal(lhs, rhs):
    return lhs == rhs


@op(
    name="timestamp-min",
    input_type={"values": types.List(types.optional(types.Timestamp()))},
    output_type=types.optional(types.Timestamp()),
)
def timestamp_min(values):
    values = [v for v in values if v != None]
    return min(values) if len(values) > 0 else None


@op(
    name="timestamp-max",
    input_type={"values": types.List(types.optional(types.Timestamp()))},
    output_type=types.optional(types.Timestamp()),
)
def timestamp_max(values):
    values = [v for v in values if v != None]
    return max(values) if len(values) > 0 else None


# Flexible date parsing, but not implemented in weaveJS so a round-trip
# is incurred to use it.
@op(render_info={"type": "function"})
def date_parse(dt_s: str) -> typing.Optional[datetime.datetime]:
    try:
        return dateutil.parser.parse(dt_s)  # type: ignore
    except dateutil.parser.ParserError:
        return None


@op(render_info={"type": "function"})
def days(days: int) -> datetime.timedelta:
    return datetime.timedelta(days=days)


# This is used as a construct, ie, it is specific and not very flexible.
# It only handles ISO formatted strings. This way we can make it match
# the js implementation exactly.
# date_parse above can be used for flexible parsing.
@op(name="timestamp", render_info={"type": "function"})
def timestamp(timestampISO: str) -> typing.Optional[datetime.datetime]:
    try:
        return datetime.datetime.fromisoformat(timestampISO)
    except ValueError:
        # TODO: Figure out why a non-iso format is getting emitted from the new DatePicker.
        # This is a hack to get around that.
        try:
            return dateutil.parser.parse(timestampISO)
        except dateutil.parser.ParserError:
            return None


@op(name="datetime-now", output_type=types.Number())
def datetime_now():
    return int(datetime.datetime.now().timestamp())



from weave import context_state as _context

_loading_builtins_token = _context.set_loading_built_ins()

from .huggingface_datasets import *
from .huggingface_models import *

from .hfmodel import ModelOutputAttention
from .model_textclassification import FullTextClassificationPipelineOutput

_context.clear_loading_built_ins(_loading_builtins_token)



__all__ = ["patch", "unpatch"]

import functools
from contextlib import contextmanager
from functools import partialmethod
from typing import Callable, Type, Union

import openai
from openai import AsyncStream, Stream
from openai.types.chat import ChatCompletion
from packaging import version

from weave import graph_client_context, run_context

from ..monitor import _get_global_monitor
from .models import *
from .util import *

old_create = openai.resources.chat.completions.Completions.create
old_async_create = openai.resources.chat.completions.AsyncCompletions.create


class partialmethod_with_self(partialmethod):
    def __get__(self, obj: Any, cls: Optional[Type[Any]] = None) -> Callable:
        return self._make_unbound_method().__get__(obj, cls)  # type: ignore


class AsyncChatCompletions:
    def __init__(self, base_create: Callable) -> None:
        self._base_create = base_create

    async def create(
        self, *args: Any, **kwargs: Any
    ) -> Union[ChatCompletion, AsyncStream[ChatCompletionChunk]]:
        if kwargs.get("stream", False):
            return self._streaming_create(*args, **kwargs)
        return await self._create(*args, **kwargs)

    async def _create(self, *args: Any, **kwargs: Any) -> ChatCompletion:
        named_args = bind_params(old_create, *args, **kwargs)
        inputs = ChatCompletionRequest.parse_obj(named_args).dict()
        with log_run("openai.chat.completions.create", inputs) as finish_run:
            result = await self._base_create(*args, **kwargs)
            finish_run(result.model_dump(exclude_unset=True))
        return result

    async def _streaming_create(
        self, *args: Any, **kwargs: Any
    ) -> AsyncStream[ChatCompletionChunk]:
        named_args = bind_params(old_create, *args, **kwargs)
        inputs = ChatCompletionRequest.parse_obj(named_args)
        with log_run(
            "openai.chat.completions.create", inputs.model_dump()
        ) as finish_run:
            # Need to put in a function so the outer function is not a
            # generator. Generators don't execute any of their body's
            # code until next() is called. But we want to create the run
            # as a child of whatever the parent is, at function call time,
            # not generator start time.
            async def _stream_create_gen():  # type: ignore
                chunks = []
                stream = await self._base_create(*args, **kwargs)
                async for chunk in stream:
                    chunks.append(chunk)
                    yield chunk
                result = reconstruct_completion(inputs.messages, chunks)  # type: ignore
                finish_run(result.model_dump(exclude_unset=True))

        return _stream_create_gen()  # type: ignore


class ChatCompletions:
    def __init__(self, base_create: Callable) -> None:
        self._base_create = base_create

    def create(
        self, *args: Any, **kwargs: Any
    ) -> Union[ChatCompletion, Stream[ChatCompletionChunk]]:
        if kwargs.get("stream", False):
            result = self._streaming_create(*args, **kwargs)
            return result
        return self._create(*args, **kwargs)

    def _create(self, *args: Any, **kwargs: Any) -> ChatCompletion:
        named_args = bind_params(old_create, *args, **kwargs)
        inputs = ChatCompletionRequest.parse_obj(named_args).dict()
        with log_run("openai.chat.completions.create", inputs) as finish_run:
            result = self._base_create(*args, **kwargs)
            finish_run(result.model_dump(exclude_unset=True))
        return result

    def _streaming_create(
        self, *args: Any, **kwargs: Any
    ) -> Stream[ChatCompletionChunk]:
        named_args = bind_params(old_create, *args, **kwargs)
        inputs = ChatCompletionRequest.parse_obj(named_args)
        with log_run(
            "openai.chat.completions.create", inputs.model_dump()
        ) as finish_run:

            def _stream_create_gen():  # type: ignore
                chunks = []
                stream = self._base_create(*args, **kwargs)
                for chunk in stream:
                    chunks.append(chunk)
                    yield chunk
                result = reconstruct_completion(inputs.messages, chunks)  # type: ignore
                finish_run(result.model_dump(exclude_unset=True))

        return _stream_create_gen()  # type: ignore


def patch() -> None:
    def _patch() -> None:
        unpatch_fqn = f"{unpatch.__module__}.{unpatch.__qualname__}()"

        if _get_global_monitor() is not None:
            # info(f"Patching OpenAI completions.  To unpatch, call {unpatch_fqn}")

            gc = graph_client_context.require_graph_client()

            hooks = ChatCompletions(old_create)
            async_hooks = AsyncChatCompletions(old_async_create)
            openai.resources.chat.completions.Completions.create = (
                partialmethod_with_self(hooks.create)
            )
            openai.resources.chat.completions.AsyncCompletions.create = (
                partialmethod_with_self(async_hooks.create)
            )

    if version.parse(openai.__version__) < version.parse("1.0.0"):
        error(
            f"this integration requires openai>=1.0.0 (got {openai.__version__}).  Please upgrade and try again"
        )
        return

    try:
        _patch()
    except Exception as e:
        error(f"problem patching: {e}, auto-unpatching")
        unpatch()
        raise Exception from e


def unpatch() -> None:
    if _get_global_monitor() is not None:
        info("Unpatching OpenAI completions")
        openai.resources.chat.completions.Completions.create = old_create
        openai.resources.chat.completions.AsyncCompletions.create = old_async_create


# TODO: centralize
@contextmanager
def log_run(call_name: str, inputs: dict[str, Any]) -> Iterator[Callable]:
    client = graph_client_context.require_graph_client()
    parent_run = run_context.get_current_run()
    # TODO: client should not need refs passed in.
    run = client.create_run(call_name, parent_run, inputs, [])

    def finish_run(output: Any) -> None:
        # TODO: client should not need refs passed in.
        client.finish_run(run, output, [])

    try:
        with run_context.current_run(run):
            yield finish_run
    except Exception as e:
        client.fail_run(run, e)
        raise



from ..decorator_op import op
from .. import weave_types as types


@op(
    name="isNone",
    input_type={"val": types.optional(types.Any())},
    output_type=types.Boolean(),
)
def is_none(val):
    return val == None



import cProfile
import os
import logging
import pathlib
import time
import traceback
import sys
import base64
import typing
import zlib
import urllib.parse
import requests
from flask import json
from werkzeug.exceptions import HTTPException

from flask import Flask, Blueprint, Response
from flask import request
from flask import abort
from flask_cors import CORS
from flask import send_from_directory, redirect
import wandb

from weave import context_state, graph, server, value_or_error
from weave import storage
from weave import registry_mem
from weave import errors
from weave import weavejs_fixes
from weave import util
from weave import engine_trace
from weave import environment
from weave import logs
from weave import filesystem
from weave.server_error_handling import client_safe_http_exceptions_as_werkzeug
from weave import storage
from weave import wandb_api
from weave.language_features.tagging import tag_store

logger = logging.getLogger(__name__)

WEAVE_CLIENT_CACHE_KEY_HEADER = "x-weave-client-cache-key"

# PROFILE_DIR = "/tmp/weave/profile"
PROFILE_DIR = None
if PROFILE_DIR is not None:
    pathlib.Path(PROFILE_DIR).mkdir(parents=True, exist_ok=True)

ERROR_STR_LIMIT = 10000

tracer = engine_trace.tracer()


def custom_dd_patch():
    import wandb_gql  # type: ignore[import]
    import wandb_graphql  # type: ignore[import]

    orig_execute = wandb_gql.Client.execute

    def execute(self, document, *args, **kwargs):
        with tracer.trace(
            "wandb_gql.Client.execute",
        ) as span:
            span.set_tag("document", wandb_graphql.print_ast(document))
            span.set_tag("variable_values", kwargs.get("variable_values", {}))
            return orig_execute(self, document, *args, **kwargs)

    wandb_gql.Client.execute = execute


if engine_trace.datadog_is_enabled():
    # Don't import this if you don't have an Agent! You'll get weird
    # crashes
    import ddtrace

    if not os.environ.get("DISABLE_WEAVE_PII"):
        ddtrace.patch_all(logging=True)
    custom_dd_patch()


# Ensure these are imported and registered
from weave import ops


# NOTE: Fixes flask dev server's auto-reload capability, by forcing it to use
# stat mode instead of watchdog mode. It turns out that "import wandb" breaks
# users of watchdog somehow. We'll need to fix that in wandb.
# A wandb fix should go out in 0.13.5.
# TODO: remove this hack when wandb 0.13.5 is released.
if os.environ.get("FLASK_DEBUG"):
    print(
        "!!! Weave server removing watchdog from sys.path for development mode. This could break other libraries"
    )
    sys.modules["watchdog.observers"] = None  # type: ignore


static_folder = os.path.join(os.path.dirname(__file__), "frontend")
blueprint = Blueprint("weave", "weave-server", static_folder=static_folder)


def import_ecosystem():
    from weave import ops
    from weave import panels
    from weave import panels_py

    # Attempt to import MVP ecosystem modules
    try:
        from weave.ecosystem import langchain, replicate
    except ImportError:
        pass

    if not util.parse_boolean_env_var("WEAVE_SERVER_DISABLE_ECOSYSTEM"):
        # try:
        #     from weave.weaveflow import faiss

        #     # Turn eager mode back off.
        #     context_state._eager_mode.set(False)
        # except (ImportError, OSError, wandb.Error):
        #     print("Error: Couldn't import faiss module for Weaveflow.")
        try:
            from weave.ecosystem import all
        except (ImportError, OSError, wandb.Error):
            pass


def log_system_info():
    WEAVE_SERVER_URL = os.environ.get("WEAVE_SERVER_URL")
    WANDB_BASE_URL = os.environ.get("WANDB_BASE_URL", "http://api.wandb.ai")
    WEAVE_LOCAL_ARTIFACT_DIR = os.environ.get("WEAVE_LOCAL_ARTIFACT_DIR")
    dot_netrc_exists = os.path.exists(os.path.expanduser("~/.netrc"))
    underscore_netrc_exists = os.path.exists(os.path.expanduser("~/_netrc"))
    WANDB_API_KEY = "REDACTED" if os.environ.get("WANDB_API_KEY") else None
    WEAVE_WANDB_COOKIE = "REDACTED" if os.environ.get("WEAVE_WANDB_COOKIE") else None
    WEAVE_WANDB_GQL_HEADERS = environment.weave_wandb_gql_headers()
    FRONTEND_ENV = json.dumps(frontend_env(), indent=2)

    logger.info("Network Config:")
    logger.info(f"  WEAVE_SERVER_URL        = {WEAVE_SERVER_URL}")
    logger.info(f"  WANDB_BASE_URL          = {WANDB_BASE_URL}")
    logger.info(f"  WEAVE_WANDB_GQL_HEADERS = {WEAVE_WANDB_GQL_HEADERS}")

    logger.info("Cache Config:")
    logger.info(f"  WEAVE_LOCAL_ARTIFACT_DIR  = {WEAVE_LOCAL_ARTIFACT_DIR}")

    logger.info("Auth Config:")
    logger.info(f"  ~/.netrc exists     = {dot_netrc_exists}")
    logger.info(f"  ~/_netrc exists     = {underscore_netrc_exists}")
    logger.info(f"  WANDB_API_KEY       = {WANDB_API_KEY}")
    logger.info(f"  WEAVE_WANDB_COOKIE  = {WEAVE_WANDB_COOKIE}")
    logger.info(f"  FRONTEND_ENV        = {FRONTEND_ENV}")


def make_app():
    logs.configure_logger()
    import_ecosystem()

    logger.info("Starting weave server")
    log_system_info()

    app = Flask(__name__)
    app.register_blueprint(blueprint)
    # Very important! We rely on key ordering on both sides!
    # Flask < 2.2 doesn't have app.json
    if hasattr(app, "json"):
        app.json.sort_keys = False
    else:
        app.config["JSON_SORT_KEYS"] = False

    CORS(app, supports_credentials=True)

    return app


class OpsCache(typing.TypedDict):
    updated_at: float
    ops_data: typing.Optional[dict]


ops_cache: typing.Optional[OpsCache] = None


@blueprint.route("/__weave/ops", methods=["GET"])
def list_ops():
    global ops_cache
    with wandb_api.from_environment():
        # TODO: this is super slow.
        if not environment.wandb_production():
            registry_mem.memory_registry.load_saved_ops()
            pass
        if (
            ops_cache is None
            or ops_cache["updated_at"] < registry_mem.memory_registry.updated_at()
        ):
            ops = registry_mem.memory_registry.list_ops()
            ret = []
            for op in ops:
                try:
                    serialized_op = op.to_dict()
                except errors.WeaveSerializeError:
                    continue
                ret.append(serialized_op)
            ops_cache = {
                "updated_at": registry_mem.memory_registry.updated_at(),
                "data": ret,
            }
    return ops_cache


class ErrorDetailsDict(typing.TypedDict):
    message: str
    error_type: str
    traceback: list[str]
    # sentry_id: typing.Union[str, None]


class ResponseDict(typing.TypedDict):
    data: list[typing.Any]
    errors: list[ErrorDetailsDict]
    node_to_error: dict[int, int]


def _exception_to_error_details(
    e: Exception, sentry_id: typing.Union[str, None]
) -> ErrorDetailsDict:
    return {
        "error_type": type(e).__name__,
        "message": str(e),
        "traceback": traceback.format_tb(e.__traceback__),
        # "sentry_id": sentry_id,
    }


def _value_or_errors_to_response(
    vore: value_or_error.ValueOrErrors,
) -> ResponseDict:
    error_lookup: dict[Exception, typing.Tuple[int, typing.Union[str, None]]] = {}
    node_errors: dict[int, int] = {}
    data: list[typing.Any] = []
    for val, error in vore.iter_items():
        if error != None:
            error = typing.cast(Exception, error)
            data.append(None)
            if error in error_lookup:
                error_ndx = error_lookup[error][0]
            else:
                sentry_id = util.capture_exception_with_sentry_if_available(error, ())
                error_ndx = len(error_lookup)
                error_lookup[error] = (error_ndx, sentry_id)
            node_errors[len(data) - 1] = error_ndx
        else:
            data.append(val)
    return {
        "data": data,
        "errors": [
            _exception_to_error_details(k, e_sentry_id)
            for k, (e_ndx, e_sentry_id) in error_lookup.items()
        ],
        "node_to_error": node_errors,
    }


def _log_errors(
    processed_response: ResponseDict, nodes: value_or_error.ValueOrErrors[graph.Node]
):
    errors: list[dict] = []

    for error in processed_response["errors"]:
        errors.append(
            {
                "message": error["message"],
                "error_type": error["error_type"],
                "traceback": error["traceback"],
                "error_tag": "node_execution_error",
                "node_strs": [],
                # "sentry_id": error["sentry_id"],
            }
        )

    for node_ndx, error_ndx in processed_response["node_to_error"].items():
        try:
            node_str = graph.node_expr_str(graph.map_const_nodes_to_x(nodes[node_ndx]))
            errors[error_ndx]["node_strs"].append(node_str)
        except Exception:
            pass

    for error_dict in errors:
        # This should be logged to DD, but 1 log per error
        # class, not 1 log per error.
        logging.error(error_dict)


def _get_client_cache_key_from_request(request):
    # Uncomment to set default to 15 second cache duration
    client_cache_key = None  # str(int(time.time() // 15))
    if WEAVE_CLIENT_CACHE_KEY_HEADER in request.headers:
        client_cache_key = request.headers[WEAVE_CLIENT_CACHE_KEY_HEADER]
    return client_cache_key


@blueprint.route("/__weave/execute", methods=["POST"])
def execute():
    """Execute endpoint used by WeaveJS."""
    with tracer.trace("read_request"):
        req_bytes = request.data
    req_compressed = zlib.compress(req_bytes)
    req_b64 = base64.b64encode(req_compressed).decode("ascii")
    logging.info(
        "Execute request (zlib): %s",
        req_b64,
    )

    if not request.json:
        abort(400, "Request body must be JSON.")
    if "graphs" not in request.json:
        abort(400, "Request body must contain a 'graphs' key.")

    # Simulate browser/server latency
    # import time
    # time.sleep(0.1)

    # use a single memartifact to serialize the entire response.
    # fixes https://weights-biases.sentry.io/issues/4022569419

    execute_args = {
        "request": request.json,
        "deref": True,
        "serialize_fn": storage.make_js_serializer(),
    }
    root_span = tracer.current_root_span()
    tag_store.record_current_tag_store_size()

    client_cache_key = _get_client_cache_key_from_request(request)

    if not PROFILE_DIR:
        start_time = time.time()
        with client_safe_http_exceptions_as_werkzeug():
            with context_state.set_client_cache_key(client_cache_key):
                response = server.handle_request(**execute_args)
        elapsed = time.time() - start_time
    else:
        # Profile the request and add a link to local snakeviz to the trace.
        profile = cProfile.Profile()
        start_time = time.time()
        try:
            with client_safe_http_exceptions_as_werkzeug():
                with context_state.set_client_cache_key(client_cache_key):
                    response = profile.runcall(server.handle_request, **execute_args)
        finally:
            elapsed = time.time() - start_time
            profile_filename = f"/tmp/weave/profile/execute.{start_time*1000:.0f}.{elapsed*1000:.0f}ms.prof"
            profile.dump_stats(profile_filename)
            if root_span:
                root_span.set_tag(
                    "profile_url",
                    "http://localhost:8080/snakeviz/"
                    + urllib.parse.quote(profile_filename),
                )

    fixed_response = response.results.safe_map(weavejs_fixes.fixup_data)
    response_payload = _value_or_errors_to_response(fixed_response)

    if root_span is not None:
        root_span.set_metric("request_size", len(req_bytes), True)
        root_span.set_metric("node_count", len(response_payload["data"]), True)
        root_span.set_metric(
            "error_count",
            len(response_payload["node_to_error"]),
            True,
        )

    _log_errors(response_payload, response.nodes)

    if request.headers.get("x-weave-include-execution-time"):
        response_payload["execution_time"] = (elapsed) * 1000

    return response_payload


@blueprint.route("/__weave/execute/v2", methods=["POST"])
def execute_v2():
    """Execute endpoint used by Weave Python"""
    # print('REQUEST', request, request.json)
    if not request.json or "graphs" not in request.json:
        abort(400)
    response = server.handle_request(request.json, deref=True)
    # print("RESPONSE BEFORE SERI", response)

    return {"data": response.results.unwrap()}


@blueprint.route("/__weave/file/<path:path>")
def send_local_file(path):
    # path is given relative to the FS root. check to see that path is a subdirectory of the
    # local artifacts path. if not, return 403. then if there is a cache scope function defined
    # call it to make sure we have access to the path
    abspath = "/" / pathlib.Path(
        path
    )  # add preceding slash as werkzeug strips this by default and it is reappended below in send_from_directory
    try:
        local_artifacts_path = pathlib.Path(filesystem.get_filesystem_dir()).absolute()
    except errors.WeaveAccessDeniedError:
        abort(403)
    if local_artifacts_path not in list(abspath.parents):
        abort(403)
    return send_from_directory("/", path)


@blueprint.before_request
def _disable_eager_mode():
    context_state._eager_mode.set(False)


def frontend_env():
    """If you add vars here, make sure to define their types in weave-js/src/config.ts"""
    return {
        "PREFIX": environment.weave_link_prefix(),
        "ANALYTICS_DISABLED": environment.analytics_disabled(),
        "ONPREM": environment.weave_onprem(),
        "WEAVE_BACKEND_HOST": environment.weave_backend_host(),
        "WANDB_BASE_URL": environment.wandb_base_url(),
        "DD_ENV": environment.dd_env(),
        "ENV_IS_CI": environment.env_is_ci(),
    }


@blueprint.route("/__frontend", defaults={"path": None})
@blueprint.route("/__frontend/<path:path>")
def frontend(path):
    """Serve the frontend with a simple fileserver over HTTP."""
    # We serve up a dynamic env.js file before all other js.
    if path is not None and path.endswith("env.js"):
        js = f"window.WEAVE_CONFIG = {json.dumps(frontend_env())}"
        return Response(js, mimetype="application/javascript")
    full_path = pathlib.Path(blueprint.static_folder) / path
    # prevent path traversal
    if not full_path.resolve().is_relative_to(blueprint.static_folder):
        return abort(403)
    if path and full_path.exists():
        return send_from_directory(blueprint.static_folder, path)
    else:
        return send_from_directory(blueprint.static_folder, "index.html")


@blueprint.route("/", defaults={"path": None})
@blueprint.route("/<path:path>")
def root_frontend(path):
    if request.args.get("unsetBetaVersion") is not None:
        resp = redirect_without_query_param("unsetBetaVersion")
        resp.set_cookie("betaVersion", "", max_age=0)
        return resp

    new_beta_version = request.args.get("betaVersion")
    if new_beta_version is not None:
        resp = redirect_without_query_param("betaVersion")
        resp.set_cookie("betaVersion", new_beta_version)
        return resp

    # To support server cases where we're mounted under an existing path, i.e.
    # /wandb/weave, then index.html will load something like /wandb/weave/.../env.js
    if path is not None:
        path = path.split("/")[-1]
    if path == "env.js":
        js = f"window.WEAVE_CONFIG = {json.dumps(frontend_env())}"
        return Response(js, mimetype="application/javascript")

    beta_version = request.cookies.get("betaVersion")
    if beta_version is not None:
        beta_version = beta_version[:9]
        content = requests.get(
            f"https://cdn.wandb.ai/weave/{beta_version}/index.html",
            stream=True,
        ).content
        return Response(content, mimetype="text/html")

    return send_from_directory(blueprint.static_folder, "index.html")


def redirect_without_query_param(param: str):
    qs_pairs = []
    for k, v in request.args.items():
        if k != param:
            qs_pairs.append(f"{k}={v}")
    qs = "&".join(qs_pairs)
    resp = redirect(f"{request.path}?{qs}")
    return resp


@blueprint.route("/__weave/hello")
def hello():
    return "hello"


@blueprint.route("/__weave/wb_viewer", methods=["POST"])
def wb_viewer():
    wandb_api.init()
    current_context = wandb_api.get_wandb_api_context()
    if not current_context:
        with wandb_api.from_environment():
            current_context = wandb_api.get_wandb_api_context()
    authenticated = current_context is not None
    user_id = None if current_context is None else current_context.user_id
    return {"authenticated": authenticated, "user_id": user_id}


app = make_app()

if os.getenv("WEAVE_SERVER_DEBUG"):

    @app.errorhandler(HTTPException)
    def handle_exception(e):
        """Return JSON instead of HTML for HTTP errors."""
        # start with the correct headers and status code from the error
        response = e.get_response()
        # replace the body with JSON
        response.data = json.dumps(
            {
                "code": e.code,
                "name": e.name,
                "description": e.description,
                "exc_info": traceback.format_exc(),
            }
        )
        response.content_type = "application/json"
        return response


if __name__ == "__main__":
    app.run(port=9994)



import weave
from .. import panel

# TODO: This id conflicts with the histogram type! Panel types
# should automatically have Panel in the name but they don't at the moment.


@weave.type("histogram")
class Histogram(panel.Panel):
    id = "histogram"



from dataclasses import dataclass
import typing
from . import weave_types as types
from . import debug_types

BT = typing.TypeVar("BT")


class OpArgs:
    NAMED_ARGS = "NAMED_ARGS"
    VAR_ARGS = "VAR_ARGS"

    @property
    def initial_arg_types(self) -> dict[str, types.Type]:
        """Returns argument types with functional types resolved"""
        raise NotImplementedError

    def weave_type(self) -> types.Type:
        raise NotImplementedError()

    def to_dict(self) -> dict:
        return {key: val.to_dict() for key, val in self.initial_arg_types.items()}

    @staticmethod
    def from_dict(_type: typing.Union[str, typing.Dict]):
        if isinstance(_type, typing.Dict):
            return OpNamedArgs(
                {
                    key: types.TypeRegistry.type_from_dict(val)
                    for key, val in _type.items()
                }
            )
        else:
            return OpVarArgs(types.TypeRegistry.type_from_dict(_type))

    def first_param_valid(self, param0_type: types.Type) -> bool:
        raise NotImplementedError

    def nonfirst_params_valid(self, param_types: list[types.Type]) -> bool:
        raise NotImplementedError

    def why_are_params_invalid(
        self, param_dict: dict[str, types.Type]
    ) -> typing.Optional[str]:
        """Attempts to assign the given param types to the op args. Returns the
        dictionary of assignment results"""
        raise NotImplementedError()

    def named_args(self) -> typing.List["NamedArg"]:
        """Retuns any named arguyments as a list"""
        raise NotImplementedError()

    def create_param_dict(self, args: list[BT], kwargs: dict[str, BT]) -> dict[str, BT]:
        """This is a key function: given a user provided list of args and kwargs, we map them
        to the current paramters and return a paramter dict that can be used by other functions."""
        raise NotImplementedError()


class OpVarArgs(OpArgs):
    kind = OpArgs.VAR_ARGS

    def __init__(self, arg_type: types.Type):
        self.arg_type = arg_type

    @property
    def initial_arg_types(self) -> dict[str, types.Type]:
        """Returns argument types with functional types resolved"""
        return {}

    def weave_type(self) -> types.Type:
        return types.Dict(types.String(), self.arg_type)

    def first_param_valid(self, param0_type: types.Type) -> bool:
        return types.optional(self.arg_type).assign_type(param0_type)

    def nonfirst_params_valid(self, param_types: list[types.Type]) -> bool:
        return all(self.arg_type.assign_type(t) for t in param_types)

    def named_args(self) -> typing.List["NamedArg"]:
        return []

    def create_param_dict(self, args: list[BT], kwargs: dict[str, BT]) -> dict[str, BT]:
        params = {f"{i}": a for i, a in enumerate(args)}
        params.update(kwargs)
        return params


class OpNamedArgs(OpArgs):
    kind = OpArgs.NAMED_ARGS

    def __init__(
        self,
        # TODO: Add callable as a type of the value
        arg_types: dict[str, types.Type],
    ):
        self.arg_types = arg_types

    def weave_type(self) -> types.TypedDict:
        return types.TypedDict(self.initial_arg_types)

    @property
    def initial_arg_types(self) -> dict[str, types.Type]:
        t: dict[str, types.Type] = {}
        for k, v in self.arg_types.items():
            if callable(v):
                t[k] = v(t)
            else:
                t[k] = v
        return t

    def first_param_valid(self, param0_type: types.Type) -> bool:
        return types.optional(list(self.arg_types.values())[0]).assign_type(param0_type)

    def nonfirst_params_valid(self, param_types: list[types.Type]) -> bool:
        arg_names = list(self.arg_types.keys())
        arg_types = list(self.arg_types.values())
        if len(param_types) < len(arg_types) - 1:
            return False
        valid_params: dict[str, types.Type] = {arg_names[0]: arg_types[0]}
        for at_key, at, pt in zip(arg_names[1:], arg_types[1:], param_types):
            if callable(at):
                at = at(valid_params)
            if not at.assign_type(pt):
                return False
            # TODO: I think this should be pt, but prior implementation has
            # as at.
            valid_params[at_key] = at
        return True

    def why_are_params_invalid(
        self, param_dict: dict[str, types.Type]
    ) -> typing.Optional[str]:
        valid_params: dict[str, types.Type] = {}
        reasons: list[str] = []
        for k, t in self.arg_types.items():
            if k not in param_dict:
                reasons.append(f"Missing parameter {k}")
                continue
            if callable(t):
                t = t(valid_params)
            if not t.assign_type(param_dict[k]):
                reasons.append(
                    f'Parameter "{k}" has invalid type\n{debug_types.why_not_assignable(t, param_dict[k])}'
                )
            valid_params[k] = t
        if reasons:
            return "\n".join(reasons)
        return None

    def named_args(self) -> typing.List["NamedArg"]:
        return [NamedArg(k, v) for k, v in self.arg_types.items()]

    def create_param_dict(self, args: list[BT], kwargs: dict[str, BT]) -> dict[str, BT]:
        # We don't actually use kwarg names!
        self_keys = list(self.arg_types.keys())
        param_keys = list(kwargs.keys())
        params: dict[str, BT] = {}
        for i, key in enumerate(self_keys):
            if i < len(args):
                params[key] = args[i]
            else:
                if i - len(args) < len(param_keys):
                    params[key] = kwargs[param_keys[i - len(args)]]
                else:
                    # This happens during dipatch
                    pass
        return params

    def __repr__(self):
        return "<OpNamedArgs %s>" % self.arg_types


@dataclass
class NamedArg:
    name: str
    type: types.Type



# W&B API Weave Ops: Runs
#
# Run projection pushdown strategy
# --------------------------------
#
# Its important for performance
# that we only select the run user data (config/summary/history) columns that
# we need. We can use the stitched graph to do that for the relevant ops, any
# pick ops connected to config/summary are used to determine the keys argument
# to the graphql field.
#
# But we also need to consider refinement. There are two different ways
# refinment will happen: explicit where there is a request for a refinement
# op like refine_summary_type in the graph, and implicit, which is all other cases.
# In the explicit case we must fetch all columns so we can produce the full
# type.
#
# The implicit case always happens, because compile refines all ops in the
# graph to fix incoming Weave0 graphs for compatibility with Weave1. But this
# refinement only needs to provide the subset of the type that will be used
# for dispatch. Ie we can compute the type of only the summary fields needed
# by the query.
#
# So the correct strategy is: if a refinement descendent is explicitly requested,
# select all fields. Otherwise, select the fields described by pick descendents.
#
# We do not yet perform projection pushdown for history. Currently to fetch
# specific history columns from the W&B API, you need to use the sampledHistory
# graphql field. But it's incorrect to make a request for one "historySpec"
# containing all the required keys, because the W&B backend will return
# only rows that have all keys present. We can instead request one spec
# for each history key, always also including the _step key, and then zip
# the results back together ourselves. If we do so we'll have to do it unsampled.
# It may be better to just move to the Runs2 strategy where Weave is directly
# responsible for scanning the data instead of using the sampledHistory field.

import json
import typing


from .. import compile_table
from ..input_provider import InputAndStitchProvider
from ..gql_op_plugin import wb_gql_op_plugin
from ..api import op
from .. import weave_types as types
from . import wb_domain_types as wdt
from ..language_features.tagging.make_tag_getter_op import make_tag_getter_op
from .wandb_domain_gql import (
    gql_prop_op,
    gql_direct_edge_op,
    gql_connection_op,
    _make_alias,
)


from . import wb_util
from .. import engine_trace
from .run_history import history_op_common


# Important to re-export ops
from .run_history import run_history_v1_legacy_ops
from .run_history import run_history_v2_parquet_media
from .run_history import run_history_v3_parquet_stream_optimized

tracer = engine_trace.tracer()

# number of rows of example data to look at to determine history type
ROW_LIMIT_FOR_TYPE_INTERROGATION = 10

# number of history key metrics to fetch from run history
LIMIT_RUN_HISTORY_KEYS = 100

# Section 1/6: Tag Getters
run_tag_getter_op = make_tag_getter_op("run", wdt.RunType, op_name="tag-run")

# Section 2/6: Root Ops
# None

# Section 3/6: Attribute Getters
gql_prop_op(
    "run-jobType",
    wdt.RunType,
    "jobType",
    types.String(),
)

gql_prop_op(
    "run-jobtype",
    wdt.RunType,
    "jobType",
    types.String(),
)

run_name = gql_prop_op(
    "run-name",
    wdt.RunType,
    "displayName",
    types.String(),
)

run_id = gql_prop_op(
    "run-internalId",
    wdt.RunType,
    "id",
    types.String(),
)

gql_prop_op(
    "run-id",
    wdt.RunType,
    "name",
    types.String(),
)

gql_prop_op(
    "run-createdAt",
    wdt.RunType,
    "createdAt",
    types.Timestamp(),
)

gql_prop_op(
    "run-defaultColorIndex",
    wdt.RunType,
    "defaultColorIndex",
    types.Number(),
)

# TODO: this should be an enum
gql_prop_op(
    "run-state",
    wdt.RunType,
    "state",
    types.String(),
)
gql_prop_op(
    "run-updatedAt",
    wdt.RunType,
    "updatedAt",
    types.Timestamp(),
)
gql_prop_op(
    "_run-historykeyinfo",
    wdt.RunType,
    "historyKeys",
    types.Dict(types.String(), types.Any()),
)

runtime = gql_prop_op(
    "run-runtime",
    wdt.RunType,
    "computeSeconds",
    types.Number(),
)

gql_prop_op(
    "run-heartbeatAt",
    wdt.RunType,
    "heartbeatAt",
    types.Timestamp(),
)

gql_prop_op(
    "run-historyLineCount",
    wdt.RunType,
    "historyLineCount",
    types.Number(),
    hidden=True,
)


def config_to_values(config: dict) -> dict:
    """
    Unfortunately config values from wandb have their data located at the .value
    property inside of the config object.
    """
    return {
        key: value["value"] if isinstance(value, dict) and "value" in value else value
        for key, value in config.items()
    }


@op(
    render_info={"type": "function"},
    hidden=True,
    plugins=wb_gql_op_plugin(lambda inputs, inner: "config"),
)
def refine_config_type(run: wdt.Run) -> types.Type:
    config_field_s = None
    try:
        # If config was explicitly requested, this will be the case.
        config_field_s = run["config"]
    except KeyError:
        # Otherwise we'll be refining implicitly in compile, but we only need
        # to provide the summary requested by the rest of the graph.
        config_field_s = run["configSubset"]
    if not config_field_s:
        config_field_s = "{}"

    return wb_util.process_run_dict_type(config_to_values(json.loads(config_field_s)))


def _make_run_config_gql_field(inputs: InputAndStitchProvider, inner: str):
    # Must be kept in sync with compile_domain:_field_selections_hardcode_merge
    stitch_obj = inputs.stitched_obj
    key_tree = compile_table.get_projection(stitch_obj)
    # we only pushdown the top level keys for now.

    top_level_keys = history_op_common.get_top_level_keys(key_tree)
    if not top_level_keys:
        # If no keys, then we must select the whole object
        return "configSubset: config"
    return f"configSubset: config(keys: {json.dumps(top_level_keys)})"


run_path_fragment = """
    project {
        id
        name
        entity {
            id
            name
        }
    }
    """


def _run_config_plugin(inputs: InputAndStitchProvider, inner: str):
    config_field = _make_run_config_gql_field(inputs, inner)
    return config_field + " " + run_path_fragment


@op(
    name="run-config",
    refine_output_type=refine_config_type,
    plugins=wb_gql_op_plugin(_run_config_plugin),
)
def config(run: wdt.Run) -> dict[str, typing.Any]:
    return wb_util.process_run_dict_obj(
        config_to_values(json.loads(run["configSubset"] or "{}")),
        wb_util.RunPath(
            run["project"]["entity"]["name"],
            run["project"]["name"],
            run["name"],
        ),
    )


@op(
    render_info={"type": "function"},
    hidden=True,
    # When refine_summary_type is explicitly requested in the graph, we ask for
    # the entire summaryMetrics field.
    plugins=wb_gql_op_plugin(lambda inputs, inner: "summaryMetrics"),
)
def refine_summary_type(run: wdt.Run) -> types.Type:
    summary_field_s = None
    try:
        # If summary was explicitly requested, this will be the case.
        summary_field_s = run["summaryMetrics"]
    except KeyError:
        # Otherwise we'll be refining implicitly in compile, but we only need
        # to provide the summary requested by the rest of the graph.
        summary_field_s = run["summaryMetricsSubset"]
    if not summary_field_s:
        summary_field_s = "{}"

    return wb_util.process_run_dict_type(json.loads(summary_field_s))


def _make_run_summary_gql_field(inputs: InputAndStitchProvider, inner: str):
    # Must be kept in sync with compile_domain:_field_selections_hardcode_merge

    stitch_obj = inputs.stitched_obj
    key_tree = compile_table.get_projection(stitch_obj)

    # we only pushdown the top level keys for now.
    top_level_keys = history_op_common.get_top_level_keys(key_tree)
    if not top_level_keys:
        # If no keys, then we must select the whole object
        return "summaryMetricsSubset: summaryMetrics"
    return f"summaryMetricsSubset: summaryMetrics(keys: {json.dumps(top_level_keys)})"


def _run_summary_plugin(inputs: InputAndStitchProvider, inner: str):
    summary_field = _make_run_summary_gql_field(inputs, inner)
    return summary_field + " " + run_path_fragment


@op(
    name="run-summary",
    refine_output_type=refine_summary_type,
    plugins=wb_gql_op_plugin(_run_summary_plugin),
)
def summary(run: wdt.Run) -> dict[str, typing.Any]:
    return wb_util.process_run_dict_obj(
        json.loads(run["summaryMetricsSubset"] or "{}"),
        wb_util.RunPath(
            run["project"]["entity"]["name"],
            run["project"]["name"],
            run["name"],
        ),
    )


def _history_as_of_plugin(inputs, inner):
    min_step = (
        inputs.raw["asOfStep"]
        if "asOfStep" in inputs.raw and inputs.raw["asOfStep"] != None
        else 0
    )
    max_step = min_step + 1
    alias = _make_alias(str(inputs.raw["asOfStep"]), prefix="history")
    return f"{alias}: history(minStep: {min_step}, maxStep: {max_step}, maxKeyLimit: {LIMIT_RUN_HISTORY_KEYS})"


def _get_history_as_of_step(
    run: wdt.Run,
    asOfStep: int,
):
    alias = _make_alias(str(asOfStep), prefix="history")

    data = run[alias]
    if isinstance(data, list):
        if len(data) > 0:
            data = data[0]
        else:
            data = None
    if data is None:
        return {}
    return json.loads(data)


@op(
    render_info={"type": "function"},
    plugins=wb_gql_op_plugin(_history_as_of_plugin),
    hidden=True,
)
def _refine_history_as_of_type(
    run: wdt.Run,
    asOfStep: int,
) -> types.Type:
    return wb_util.process_run_dict_type(_get_history_as_of_step(run, asOfStep))


@op(
    name="run-historyAsOf",
    refine_output_type=_refine_history_as_of_type,
    plugins=wb_gql_op_plugin(_history_as_of_plugin),
)
def history_as_of(
    run: wdt.Run,
    asOfStep: int,
) -> dict[str, typing.Any]:
    return _get_history_as_of_step(run, asOfStep)


# Section 4/6: Direct Relationship Ops
gql_direct_edge_op(
    "run-user",
    wdt.RunType,
    "user",
    wdt.UserType,
)

gql_direct_edge_op(
    "run-project",
    wdt.RunType,
    "project",
    wdt.ProjectType,
)

# Section 5/6: Connection Ops
gql_connection_op(
    "run-usedArtifactVersions",
    wdt.RunType,
    "inputArtifacts",
    wdt.ArtifactVersionType,
    {},
    lambda inputs: "first: 100",
)

gql_connection_op(
    "run-loggedArtifactVersions",
    wdt.RunType,
    "outputArtifacts",
    wdt.ArtifactVersionType,
    {},
    lambda inputs: "first: 100",
)


# Section 6/6: Non Standard Business Logic Ops
@op(
    name="run-link",
    plugins=wb_gql_op_plugin(
        lambda inputs, inner: """
    displayName
    project {
        id
        name
        entity {
            id
            name
        }
    }
"""
    ),
)
def link(run: wdt.Run) -> wdt.Link:
    return wdt.Link(
        run["displayName"],
        f'/{run["project"]["entity"]["name"]}/{run["project"]["name"]}/runs/{run["name"]}',
    )


@op(
    name="constructor-wbRunLink",
    input_type={
        "entity_name": types.optional(types.String()),
        "project_name": types.optional(types.String()),
        "name": types.optional(types.String()),
    },
)
def str_run_link(entity_name, project_name, name) -> wdt.Link:
    return wdt.Link(
        f"{entity_name}/{project_name}/{name}",
        f"/{entity_name}/{project_name}/runs/{name}/overview",
    )


def run_logged_artifact_version_gql_plugin(inputs, inner):
    artifact_name = inputs.raw["artifactVersionName"]
    alias = _make_alias(artifact_name, prefix="artifact")
    if ":" not in artifact_name:
        artifact_name += ":latest"
    artifact_name = json.dumps(artifact_name)
    return f"""
    project {{
        {alias}: artifact(name: {artifact_name}) {{
            {wdt.ArtifactVersion.REQUIRED_FRAGMENT}
            {inner}
        }}
    }}"""


@op(
    name="run-loggedArtifactVersion",
    plugins=wb_gql_op_plugin(run_logged_artifact_version_gql_plugin),
)
def run_logged_artifact_version(
    run: wdt.Run, artifactVersionName: str
) -> wdt.ArtifactVersion:
    alias = _make_alias(artifactVersionName, prefix="artifact")
    return wdt.ArtifactVersion.from_keys(run["project"][alias])



"""Pydoc browsing.

Improvements we could make:
- Give methods chainable names (by using @weave_class and methods for auto-naming)
- Fix inability to use .name() ops
- Add more stuff
- Access attributes with getattr instead of declaring named methods.
"""

import weave
import types
import inspect


class PyModule(weave.types.Type):
    instance_classes = types.ModuleType

    def instance_to_dict(self, obj):
        return {"module_name": obj.__name__}

    def instance_from_dict(self, d):
        return __import__(d["module_name"])


class PyClass(weave.types.Type):
    # TODO: Will registering this break stuff? Everything is type.
    instance_classes = type

    def instance_to_dict(self, obj):
        return {"module_name": obj.__module__.__name__, "class_name": obj.__name__}

    def instance_from_dict(self, d):
        return getattr(__import__(d["module_name"]), d["class_name"])


class PyFunction(weave.types.Type):
    # TODO: Will registering this break stuff? Everything is type.
    instance_classes = types.FunctionType

    def instance_to_dict(self, obj):
        return {"module_name": obj.__module__.__name__, "function_name": obj.__name__}

    def instance_from_dict(self, d):
        return getattr(__import__(d["module_name"]), d["function_name"])


@weave.op()
def module_name(module: types.ModuleType) -> str:
    return module.__name__


@weave.op()
def module_doc(module: types.ModuleType) -> weave.ops.Markdown:
    return weave.ops.Markdown(module.__doc__ or "")


@weave.op()
def module_classes(module: types.ModuleType) -> list[type]:
    return [m[1] for m in inspect.getmembers(module, predicate=inspect.isclass)]


@weave.op()
def module_class(module: types.ModuleType, class_name: str) -> type:
    # TODO: type check? or have __getattr__ work instead and provide
    # refinement
    return getattr(module, class_name)


@weave.op()
def module_functions(module: types.ModuleType) -> list[types.FunctionType]:
    def is_func(m):
        return inspect.isfunction(m) or inspect.isbuiltin(m)

    return [m[1] for m in inspect.getmembers(module, predicate=is_func)]


@weave.op()
def module_function(module: types.ModuleType, function_name: str) -> types.FunctionType:
    # TODO: type check? or have __getattr__ work instead and provide
    # refinement
    return getattr(module, function_name)


@weave.op(render_info={"type": "function"})
def pyclass(module_name: str, class_name: str) -> type:
    return getattr(__import__(module_name), class_name)


@weave.op()
def pyclass_module(pyclass: type) -> types.ModuleType:
    return __import__(pyclass.__module__)


@weave.op()
def pyclass_doc(pyclass: type) -> weave.ops.Markdown:
    return weave.ops.Markdown(pyclass.__doc__ or "")


@weave.op()
def class_name(pyclass: type) -> str:
    return pyclass.__name__


@weave.op()
def class_methods(pyclass: type) -> list[types.FunctionType]:
    def is_func(m):
        return inspect.ismethod(m) or inspect.isfunction(m)

    return [m[1] for m in inspect.getmembers(pyclass, predicate=is_func)]


@weave.op()
def class_method(pyclass: type, method_name: str) -> types.FunctionType:
    return getattr(pyclass, method_name)


@weave.op()
def pyfunction(module_name: str, function_name: str) -> types.FunctionType:
    return getattr(__import__(module_name), function_name)


@weave.op()
def function_name(pyfunction: types.FunctionType) -> str:
    return pyfunction.__name__


@weave.op()
def function_doc(pyfunction: types.FunctionType) -> weave.ops.Markdown:
    return weave.ops.Markdown(pyfunction.__doc__ or "")


@weave.type()
class ModulePanel(weave.Panel):
    id = "ModulePanel"
    input_node: weave.Node[types.ModuleType]

    @weave.op()
    def render(self) -> weave.panels.Card:
        module = self.input_node
        return weave.panels.Card(
            title=module.module_name(),  # type: ignore
            subtitle="python module",
            content=[
                weave.panels.CardTab(
                    name="Description",
                    content=weave.panels.PanelMarkdown(module.module_doc()),  # type: ignore
                ),
                weave.panels.CardTab(
                    name="Classes",
                    content=weave.panels.Table(
                        module.module_classes(),  # type: ignore
                        columns=[
                            lambda c: weave.panels.WeaveLink(
                                c.class_name(),
                                to=lambda inp: module.module_class(inp),  # type: ignore
                            )
                        ],
                    ),
                ),
                weave.panels.CardTab(
                    name="Functions",
                    content=weave.panels.Table(
                        module.module_functions(),  # type: ignore
                        columns=[
                            lambda c: weave.panels.WeaveLink(
                                c.function_name(),
                                to=lambda inp: module.module_function(inp),  # type: ignore
                            )
                        ],
                    ),
                ),
            ],
        )


@weave.type()
class ClassPanel(weave.Panel):
    id = "ClassPanel"
    input_node: weave.Node[type]

    @weave.op()
    def render(self) -> weave.panels.Card:
        cls = self.input_node
        return weave.panels.Card(
            title=cls.class_name(),  # type: ignore
            subtitle="python class",
            content=[
                weave.panels.CardTab(
                    name="Description",
                    content=weave.panels.PanelMarkdown(cls.pyclass_doc()),  # type: ignore
                ),
                weave.panels.CardTab(
                    name="Methods",
                    content=weave.panels.Table(
                        cls.class_methods(),  # type: ignore
                        columns=[
                            lambda m: weave.panels.WeaveLink(
                                m.function_name(),
                                to=lambda inp: cls.class_method(inp),  # type: ignore
                            )
                        ],
                    ),
                ),
            ],
        )


@weave.type()
class FunctionPanel(weave.Panel):
    id = "FunctionPanel"
    input_node: weave.Node[types.FunctionType]

    @weave.op()
    def render(self) -> weave.panels.Card:
        func = self.input_node
        return weave.panels.Card(
            title=func.function_name(),  # type: ignore
            subtitle="python function",
            content=[
                weave.panels.CardTab(
                    name="Description",
                    content=weave.panels.PanelMarkdown(func.function_doc()),  # type: ignore
                ),
            ],
        )



import json
import typing

from ... import graph
from ... import compile
from ... import op_args
from ... import registry_mem
from ... import weave_types as types
from .. import wb_domain_types as wdt
from ...ops_primitives import make_list
from ...ops_arrow.list_ops import concat
from ...ops_arrow import ArrowWeaveList
from ...arrow.concat import concatenate_all
from ... import util
from ... import errors
from ... import io_service
from ...mappers_arrow import map_to_arrow
from ... import engine_trace
from ...compile_domain import InputAndStitchProvider
from ... import compile_table
from ... import artifact_base
from ...language_features.tagging.tagged_value_type import TaggedValueType

from ...api import use

import pyarrow as pa
from pyarrow import parquet as pq


from .. import wb_util
from .. import table

from ... import artifact_fs
from ...wandb_interface import wandb_stream_table
from ...compile_table import KeyTree
from ... import _dict_utils
from ... import gql_json_cache

tracer = engine_trace.tracer()


class TypeCount(typing.TypedDict):
    type: typing.Optional[str]
    count: int
    keys: dict[str, list["TypeCount"]]  # type: ignore
    items: list["TypeCount"]  # type: ignore
    nested_types: list[str]


def history_key_type_count_to_weave_type(tc: TypeCount) -> types.Type:
    from ..wbmedia import ImageArtifactFileRefType
    from ..trace_tree import WBTraceTree

    tc_type = tc["type"]
    if tc_type == "string":
        return types.String()
    elif tc_type == "number":
        return types.Number()
    elif tc_type == "nil":
        return types.NoneType()
    elif tc_type == "bool":
        return types.Boolean()
    elif tc_type == "map":
        keys = tc["keys"] if "keys" in tc else {}
        return types.TypedDict(
            {
                key: types.union(
                    *[history_key_type_count_to_weave_type(vv) for vv in val]
                )
                for key, val in keys.items()
            }
        )
    elif tc_type == "list":
        if "items" not in tc:
            return types.List(types.UnknownType())
        return types.List(
            types.union(*[history_key_type_count_to_weave_type(v) for v in tc["items"]])
        )
    elif tc_type == "histogram":
        return wb_util.WbHistogram.WeaveType()  # type: ignore
    elif tc_type == "table-file":
        extension = types.Const(types.String(), "json")
        return artifact_fs.FilesystemArtifactFileType(extension, table.TableType())
    elif tc_type == "joined-table":
        extension = types.Const(types.String(), "json")
        return artifact_fs.FilesystemArtifactFileType(
            extension, table.JoinedTableType()
        )
    elif tc_type == "partitioned-table":
        extension = types.Const(types.String(), "json")
        return artifact_fs.FilesystemArtifactFileType(
            extension, table.PartitionedTableType()
        )
    elif tc_type == "image-file":
        return ImageArtifactFileRefType()
    elif tc_type == "wb_trace_tree":
        return WBTraceTree.WeaveType()  # type: ignore
    elif tc_type == "images/separated":
        return types.List(ImageArtifactFileRefType())
    elif isinstance(tc_type, str):
        possible_type = wandb_stream_table.maybe_history_type_to_weave_type(tc_type)
        if possible_type is not None:
            return possible_type

    # Hack: there are circumstances where the user has logged data with a
    # `_type` key. This is nasty... because so much of our logic depends on
    # `_type` being a reserved key. In such circumstances the gorilla history
    # parser will still return the correct keys. This is essentially the same as
    # the `map` case above. This MUST go last however, since we want to properly
    # catch any named types first.
    if "keys" in tc:
        return types.TypedDict(
            {
                key: types.union(
                    *[history_key_type_count_to_weave_type(vv) for vv in val]
                )
                for key, val in tc["keys"].items()
            }
        )
    return types.UnknownType()


def get_top_level_keys(key_tree: KeyTree) -> list[str]:
    top_level_keys = list(
        map(
            _dict_utils.unescape_dots,
            set(
                next(iter(_dict_utils.split_escaped_string(key)))
                for key in key_tree.keys()
            ),
        )
    )
    return top_level_keys


def get_full_columns(columns: typing.Optional[list[str]]):
    if columns is None:
        return None
    return list(set([*columns, "_step"]))


def get_full_columns_prefixed(run: wdt.Run, columns: typing.Optional[list[str]]):
    all_columns = get_full_columns(columns)
    all_paths = list(
        gql_json_cache.use_json(run.get("historyKeys", "{}")).get("keys", {}).keys()
    )
    return _filter_known_paths_to_requested_paths(all_paths, all_columns)


def flatten_typed_dicts(type_: types.TypedDict) -> list[str]:
    props = type_.property_types
    paths = []
    for key, val in props.items():
        _, non_none_val_type = types.split_none(val)
        # The prior code did an assignment check like this:
        #   if types.optional(types.TypedDict({})).assign_type(
        #       val
        #   ) and not types.NoneType().assign_type(val):
        #   However that doesn't work because Ref<TypedDict<A>> is assignable
        #   to TypedDict<A>
        # Do we care about a Union of TypedDict? I don't think so, since we're
        # operating on ArrowWeaveList here, which always does merge types and so can only
        # have a union with a single TypedDict in it
        if isinstance(non_none_val_type, types.TypedDict):
            subpaths = flatten_typed_dicts(typing.cast(types.TypedDict, val))
            paths += [f"{key}.{path}" for path in subpaths]
        else:
            paths.append(key)
    return list(set(paths))


def _without_tags(type_: types.Type) -> types.Type:
    if isinstance(type_, TaggedValueType):
        return _without_tags(type_.value)
    return type_


def _is_list_like(type_: types.Type) -> bool:
    return hasattr(type_, "object_type")


def _is_dict_like(type_: types.Type) -> bool:
    return hasattr(type_, "property_types")


def _get_key_typed_node(node: graph.Node) -> graph.Node:
    with tracer.trace("run_history_inner_refine") as span:
        # This block will not normally be needed. But in some cases (for example lambdas)
        # we will have to refine the node to get the history type.
        if not isinstance(node, graph.OutputNode):
            raise errors.WeaveWBHistoryTranslationError(
                f"Expected output node, found {type(node)}"
            )
        op = registry_mem.memory_registry._ops[node.from_op.name]
        with compile.enable_compile():
            if not isinstance(op.input_type, op_args.OpNamedArgs):
                raise errors.WeaveWBHistoryTranslationError(
                    f"Expected named args, found {type(op.input_type)}"
                )
            return op.lazy_call(
                **{
                    k: n
                    for k, n in zip(
                        op.input_type.arg_types, node.from_op.inputs.values()
                    )
                }
            )


def _filter_known_paths_to_requested_paths(
    known_paths: list[str], requested_paths: list[str]
) -> list[str]:
    history_cols = []
    for path in known_paths:
        for requested_path in requested_paths:
            if path == requested_path or path.startswith(requested_path + "."):
                history_cols.append(path)
    return history_cols


def _history_node_known_keys(node: graph.Node) -> list[str]:
    node_type = node.type

    # First, we expect the type to be list like
    if not _is_list_like(node_type):
        return []

    node_type = typing.cast(types.List, node_type)
    node_type = node_type.object_type

    # Next, it is possible we have one more layer of list (if the
    # node was mapped)
    if _is_list_like(node_type):
        node_type = typing.cast(types.List, node_type)
        node_type = node_type.object_type

    # Finally, we would expect the node to be dict like and have
    # at least one property.
    if _is_dict_like(node_type):
        node_type = typing.cast(types.TypedDict, node_type)
        if len(node_type.property_types) > 0:
            node_type = typing.cast(types.TypedDict, _without_tags(node_type))
            return flatten_typed_dicts(node_type)

    return []


def make_run_history_gql_field(inputs: InputAndStitchProvider, inner: str):
    # Must be kept in sync with compile_domain:_field_selections_hardcode_merge
    # I moved the column pushdown before the gql step, so we actually have the
    # columns directly available here.... could be possible we need to revisit
    top_level_keys = inputs.raw.get("history_cols")
    if top_level_keys == None:
        stitch_obj = inputs.stitched_obj
        key_tree = compile_table.get_projection(stitch_obj)

        # we only pushdown the top level keys for now.
        top_level_keys = get_top_level_keys(key_tree)

    if not top_level_keys:
        # If no keys, then we cowardly refuse to blindly fetch entire history table
        return "historyKeys"

    # We need to figure out the known history keys. In the vast majority of
    # situations, the node's type will contain this information (by virtue of
    # the refinement step during compilation. There is an edge case with nested
    # lambdas where this will not be the case, and we need to explicitly call
    # `_get_key_typed_node` to get the correctly typed node. Moreover, we need
    # to properly handle mapping here. The fallback is to just use the `_step`
    # key.
    all_known_paths = ["_step"]

    node = inputs.stitched_obj.node
    paths_from_node = _history_node_known_keys(inputs.stitched_obj.node)

    # If we don't have any paths from the node, then we need to refine it.
    if len(paths_from_node) == 0:
        node = _get_key_typed_node(node)
        paths_from_node = _history_node_known_keys(node)

    all_known_paths += paths_from_node

    history_cols = _filter_known_paths_to_requested_paths(
        list(set(all_known_paths)), get_full_columns(top_level_keys)
    )

    project_fragment = """
        project {
        id
        name
        entity {
            id
            name
        }
    }
    """

    return f"""
    historyKeys
    sampledParquetHistory: parquetHistory(liveKeys: {json.dumps(history_cols)}) {{
        liveData
        parquetUrls
    }}
    {project_fragment}
    """


def refine_history_type(
    run: wdt.Run,
    columns: typing.Optional[list[str]] = None,
) -> types.TypedDict:
    if "historyKeys" not in run.keys:
        raise ValueError("historyKeys not in run gql")

    historyKeys = gql_json_cache.use_json(run["historyKeys"])["keys"]

    return _refine_history_type_inner(historyKeys, columns)


def _refine_history_type_inner(
    historyKeys: dict[str, dict],
    columns: typing.Optional[list[str]] = None,
) -> types.TypedDict:
    prop_types: dict[str, types.Type] = {}
    for key, key_details in historyKeys.items():
        if key.startswith("system/") or (columns is not None and key not in columns):
            # skip system metrics for now
            continue

        type_counts: list[TypeCount] = key_details["typeCounts"]
        wt = types.merge_many_types(
            [history_key_type_count_to_weave_type(tc) for tc in type_counts]
        )

        if wt == types.UnknownType():
            util.capture_exception_with_sentry_if_available(
                errors.WeaveTypeWarning(
                    f"Unable to determine history key type for key {key} with types {type_counts}"
                ),
                (str([tc["type"] for tc in type_counts]),),
            )
            wt = types.NoneType()

        # _step is a special key that is always guaranteed to be a nonnull number.
        # other keys may be undefined at particular steps so we make them optional.
        if key in ["_step", "_runtime", "_timestamp"]:
            prop_types[key] = wt
        else:
            prop_types[key] = types.optional(wt)

    return types.TypedDict(prop_types)


def history_keys(run: wdt.Run) -> list[str]:
    if "historyKeys" not in run.keys:
        raise ValueError("historyKeys not in run gql")
    if "sampledParquetHistory" not in run.keys:
        raise ValueError("sampledParquetHistory not in run gql")
    object_type = refine_history_type(run)

    return list(object_type.property_types.keys())


def awl_from_local_parquet_path(
    path: str,
    object_type: typing.Optional[types.TypedDict],
    columns: list[str] = [],
    artifact: typing.Optional[artifact_base.Artifact] = None,
) -> ArrowWeaveList:
    with tracer.trace("pq.read_metadata") as span:
        span.set_tag("path", path)
        meta = pq.read_metadata(path)
    file_schema = meta.schema
    columns_to_read = [c for c in columns if c in file_schema.to_arrow_schema().names]
    with tracer.trace("pq.read_table") as span:
        span.set_tag("path", path)
        table = pq.read_table(path, columns=columns_to_read)

    # convert table to ArrowWeaveList
    with tracer.trace("make_awl") as span:
        awl: ArrowWeaveList = ArrowWeaveList(
            table, object_type=object_type, artifact=artifact
        )
    return awl


def process_history_awl_tables(tables: list[ArrowWeaveList]):
    concatted = concat_awls(tables)
    if isinstance(concatted, ArrowWeaveList):
        parquet_history = awl_to_pa_table(concatted)
    else:
        # empty table
        return None

    return sort_history_pa_table(parquet_history)


def concat_awls(awls: list[ArrowWeaveList]):
    return concatenate_all(awls)


def awl_to_pa_table(awl: ArrowWeaveList):
    rb = pa.RecordBatch.from_struct_array(
        awl._arrow_data
    )  # this pivots to columnar layout
    return pa.Table.from_batches([rb])


def sort_history_pa_table(table: pa.Table):
    with tracer.trace("pq.sort"):
        table_sorted_indices = pa.compute.bottom_k_unstable(
            table, sort_keys=["_step"], k=len(table)
        )

    with tracer.trace("pq.take"):
        return table.take(table_sorted_indices)


def read_history_parquet(run: wdt.Run, columns=None):
    io = io_service.get_sync_client()
    object_type = refine_history_type(run, columns=columns)
    tables = []
    for url in run["sampledParquetHistory"]["parquetUrls"]:
        local_path = io.ensure_file_downloaded(url)
        if local_path is not None:
            path = io.fs.path(local_path)
            awl = awl_from_local_parquet_path(path, object_type, columns=columns)
            tables.append(awl)
    if len(tables) == 0:
        return None
    return process_history_awl_tables(tables)


def mock_history_rows(
    run: wdt.Run, use_arrow: bool = True
) -> typing.Union[ArrowWeaveList, list]:
    # we are in the case where we have blindly requested the entire history object.
    # we refuse to fetch that, so instead we will just inspect the historyKeys and return
    # a dummy history object that can bte used as a proxy for downstream ops (e.g., count).

    step_type = types.TypedDict({"_step": types.Int()})
    steps: typing.Union[ArrowWeaveList, list] = []
    history_keys = gql_json_cache.use_json(run["historyKeys"])

    last_step = history_keys["lastStep"]
    keys = history_keys["keys"]
    for key, key_details in keys.items():
        if key == "_step":
            type_counts: list[TypeCount] = key_details["typeCounts"]
            count = type_counts[0]["count"]
            # generate fake steps
            steps = [{"_step": i} for i in range(count)]
            steps[-1]["_step"] = last_step
            assert len(steps) == count
            break

    if use_arrow:
        mapper = map_to_arrow(step_type, None, [])
        result = pa.array(steps, type=mapper.result_type())
        steps = ArrowWeaveList(result, step_type)

    return steps



import dataclasses
import typing

import weave

from . import weave_plotly
from ... import weave_internal
from ... import panel_util


@weave.type()
class DistributionConfig:
    # value_fn and label_fn are functions of input item. They will be
    # represented in the config as Const(FunctionType(...), Node).
    # We should make a better type to represent this, so it can be
    # distinguished from an expression like bin_size.
    value_fn: weave.Node[typing.Any] = dataclasses.field(
        default_factory=lambda: weave.graph.VoidNode()
    )
    label_fn: weave.Node[typing.Any] = dataclasses.field(
        default_factory=lambda: weave.graph.VoidNode()
    )
    # This is an expression. It will be stored in the config as Node.
    bin_size: weave.Node[typing.Any] = dataclasses.field(
        default_factory=lambda: weave.graph.VoidNode()
    )


# The interface for constructing this Panel from Python
@weave.type()
class Distribution(weave.Panel):
    id = "Distribution"
    input_node: weave.Node[list[typing.Any]]
    config: typing.Optional[DistributionConfig] = None

    @weave.op()
    def initialize(self) -> DistributionConfig:
        input_node = self.input_node
        unnested = weave.ops.unnest(input_node)
        return DistributionConfig(
            value_fn=weave_internal.define_fn(
                {"item": unnested.type.object_type}, lambda item: item  # type: ignore
            ),
            label_fn=weave_internal.define_fn(
                {"item": unnested.type.object_type}, lambda item: item  # type: ignore
            ),
            # Would be nice to call this weave.expr
            bin_size=panel_util.make_node(10),  # type: ignore
        )

    @weave.op()
    def render_config(self) -> weave.panels.Group:
        config = typing.cast(DistributionConfig, self.config)
        return weave.panels.Group(
            items={
                "value_fn": weave.panels.LabeledItem(
                    label="value",
                    item=weave.panels.FunctionEditor(config.value_fn),
                ),
                "label_fn": weave.panels.LabeledItem(
                    label="label",
                    item=weave.panels.FunctionEditor(config.label_fn),
                ),
                "bin_size": weave.panels.LabeledItem(
                    label="bin_size",
                    # Must execute here because bin_size is an expression.
                    # Editor panels take the actual type they edit (in this
                    # case float). The slider will edit the value of the
                    # const node directly in the config, or if the expression
                    # refers to variables, the edit will be routed to the appropriate
                    # owner.
                    item=weave.panels.Slider(config.bin_size.execute()),  # type: ignore
                ),
            }
        )

    # The render op. This renders the panel.
    @weave.op()
    def render(self) -> weave_plotly.PanelPlotly:
        input_node = self.input_node
        config = typing.cast(DistributionConfig, self.config)

        if not weave.types.union(
            weave.types.NoneType(), weave.types.String(), weave.types.Float()
        ).assign_type(
            config.value_fn.type.output_type  # type: ignore
        ):
            # TODO: need a nicer way to return error states
            return weave.panels.PanelString("Invalid value_fn")  # type: ignore
        # We always unnest, so that we can compare across groups of items
        # easily. (the Distribution notebook)
        unnested = weave.ops.unnest(input_node)
        bin_size = config.bin_size

        def bin_func(item):
            value_fn_output_type = config.value_fn.type.output_type
            label_fn_output_type = config.label_fn.type.output_type
            group_items = {}
            if value_fn_output_type == weave.types.String():
                group_items["value"] = config.value_fn(item)
            else:
                group_items["value"] = (
                    round(config.value_fn(item) / bin_size) * bin_size
                )

            if not weave.types.optional(weave.types.String()).assign_type(
                label_fn_output_type
            ):
                group_items["label"] = "no_label"
            else:
                group_items["label"] = config.label_fn(item)

            return weave.ops.dict_(**group_items)

        binned = unnested.groupby(lambda item: bin_func(item)).map(
            lambda group: weave.ops.dict_(
                value=group.groupkey()["value"],
                label=group.groupkey()["label"],
                count=group.count(),
            )
        )
        fig = weave_plotly.plotly_barplot(binned)
        return weave_plotly.PanelPlotly(fig)


# PanelPlot version... This is used by a test.
@weave.op()
def distribution_panel_plot_render(
    input_node: weave.Node[list[typing.Any]], config: DistributionConfig
) -> weave.panels.Plot:
    unnested = weave.ops.unnest(input_node)
    bin_size = weave.ops.execute(config.bin_size)

    def bin_func(item):
        value_fn_output_type = config.value_fn.type.output_type
        label_fn_output_type = config.label_fn.type.output_type
        group_items = {}
        if value_fn_output_type == weave.types.String():
            group_items["value"] = config.value_fn(item)
        else:
            group_items["value"] = round(config.value_fn(item) / bin_size) * bin_size

        if label_fn_output_type == weave.types.Invalid():
            group_items["label"] = "no_label"
        else:
            group_items["label"] = config.label_fn(item)

        res = weave.ops.dict_(**group_items)
        return res

    binned = unnested.groupby(lambda item: bin_func(item)).map(
        lambda group: weave.ops.dict_(
            value=group.groupkey()["value"],
            label=group.groupkey()["label"],
            count=group.count(),
        )
    )

    return weave.panels.Plot(
        binned,
        x=lambda row: row["value"],
        y=lambda row: row["count"],
        label=lambda row: row["label"],
        mark="bar",
    )



from typing import Tuple, Union, Callable, Any, TypeVar, Type, Optional

_T = TypeVar("_T")

# This adds VSCode/pylance/pyright typechecking support for weave.obj() using
# Data Class Transforms. See https://peps.python.org/pep-0681/.
def __dataclass_transform__(
    *,
    eq_default: bool = True,
    order_default: bool = False,
    kw_only_default: bool = False,
    field_descriptors: Tuple[Union[Type, Callable[..., Any]], ...] = (()),
) -> Callable[[_T], _T]: ...
@__dataclass_transform__()
def type(
    __override_name: Optional[str] = None,
    __is_simple: bool = False,
    __init: Optional[bool] = True,
    __mixins: Optional[list[Type]] = None,
) -> Callable[[_T], _T]: ...



import weave

from .. import val_const
from .. import ops_primitives


def test_difference_type():
    diff = ops_primitives.difference(
        [val_const.const("a"), val_const.const("b")],
        [val_const.const("b")],
    )
    # When difference operations on Const Union types, we get a const union
    # type back with the difference applied (ie difference works at type level
    # when types are available as Consts)
    assert diff.type == weave.types.List(
        weave.types.union(weave.types.Const(weave.types.String(), "a"))
    )



import typing
import contextlib
import datetime
import logging
import shutil
import time
import os

from . import context_state
from . import engine_trace
from . import environment
from . import errors
from . import wandb_api

statsd = engine_trace.statsd()  # type: ignore
logger = logging.getLogger("root")

"""
Returns a timestamp bucketed by interval days that is calculated from epoch time. For example, if we want
a bucket interval of 1 day, the anytime this function is called between 0 and 86399 seconds during the first
day of unix time (Jan 1 1970), it will return Jan 2 1970. This function returns the END time of the bucket
so it is easier for the caller to know when a cache interval will no longer be used. 

For another example, if we want a bucket interval of 7 days, then here are some example values:
system time             bucketed time
11:22 Jan 1 1970       00:00 Jan 8 1970
23:59 Jan 7 1970       00:00 Jan 14 1970
00:00 Jan 8 1970       00:00 Jan 15 1970
16:00 Jan 12 1970      00:00 Jan 19 1970
12:00 Jan 3 2024       00:00 Jan 10 2024
"""


def bucket_timestamp(interval_days: int) -> str:
    if interval_days == 0:
        return "0"
    now = time.time()
    interval_seconds = interval_days * 24 * 60 * 60
    num_intervals = int(now / interval_seconds)
    # use the bucket end time because this makes it easy to know when a cache interval will no longer be used
    bucket_end_time = (num_intervals + 1) * interval_seconds
    return str(
        datetime.datetime.fromtimestamp(bucket_end_time, datetime.timezone.utc)
        .timestamp()
        .__floor__()
    )


@contextlib.contextmanager
def time_interval_cache_prefix(
    cache_prefix_input: typing.Optional[str] = None,
) -> typing.Generator[None, None, None]:
    cache_prefix = cache_prefix_input
    if cache_prefix is None:
        duration = environment.cache_duration_days()
        if duration != 0:
            cache_prefix = bucket_timestamp(duration)
            logger.info(f"New cache prefix {cache_prefix}")
    token = context_state._cache_prefix_context.set(cache_prefix)
    try:
        yield
    finally:
        context_state._cache_prefix_context.reset(token)


def get_cache_prefix_context() -> typing.Optional[str]:
    return context_state._cache_prefix_context.get()


def clear_cache() -> None:
    # Read the directory address and threshold from the environment variable
    directory_path = environment.weave_filesystem_dir()

    now = datetime.datetime.now().timestamp()
    # buffer is in seconds
    buffer = 60 * 60 * 24 * environment.cache_deletion_buffer_days()  # days of buffer

    # Validate the directory path
    if not directory_path:
        logging.info("WEAVE_PYTHON_CACHE is not set.")
        return
    if not os.path.isdir(directory_path):
        logging.info(f"{directory_path} is not a valid directory.")
        return

    # for each cache in the directory, check if its expired past the buffer
    for item in os.listdir(directory_path):
        try:
            item_timestamp = int(item)
            item_path = os.path.join(directory_path, item)

            # If the item is expired and we are past the buffer, delete it
            if item_timestamp + buffer < now:
                # Delete the data
                shutil.rmtree(item_path)
                logging.info(f"Deleted {item}.")
        except:
            logging.info(f"Error deleting {item}.")
            continue


def get_user_cache_key() -> typing.Optional[str]:
    ctx = wandb_api.get_wandb_api_context()
    if ctx is None:
        if environment.is_public():
            raise errors.WeaveAccessDeniedError("No user set in public environment")
        return None
    return ctx.user_id


# TODO: stats, test

CacheKeyType = typing.TypeVar("CacheKeyType")
CacheValueType = typing.TypeVar("CacheValueType")


class LruTimeWindowCache(typing.Generic[CacheKeyType, CacheValueType]):
    """A cache that stores values for a fixed amount of time.

    Respects the user cache key, so that different users don't share the same cache.
    """

    class NotFound:
        pass

    NOT_FOUND = NotFound()

    def __init__(
        self,
        max_age: datetime.timedelta,
        now_fn: typing.Callable[[], datetime.datetime] = datetime.datetime.now,
    ) -> None:
        self.max_age = max_age
        self._now_fn = now_fn

        # Items are time ordered, with oldest at the front.
        self._cache: dict[
            typing.Tuple[typing.Optional[str], CacheKeyType],
            typing.Tuple[datetime.datetime, CacheValueType],
        ] = {}

    def _full_key(
        self, key: CacheKeyType
    ) -> typing.Tuple[typing.Optional[str], CacheKeyType]:
        return (get_user_cache_key(), key)

    def _prune(self, now: datetime.datetime) -> None:
        for key, val in list(self._cache.items()):
            if now - val[0] > self.max_age:
                del self._cache[key]
            else:
                break
        statsd.gauge("weave.cache.size", len(self._cache))

    def get(self, key: CacheKeyType) -> typing.Union[NotFound, CacheValueType]:
        full_key = self._full_key(key)
        val = self._cache.get(full_key)
        if val is None:
            statsd.increment("weave.cache.miss")
            return self.NOT_FOUND
        # Set the value again to move it to the end of the cache
        self.set(key, val[1])
        statsd.increment("weave.cache.hit")
        return val[1]

    def set(self, key: CacheKeyType, value: CacheValueType) -> None:
        full_key = self._full_key(key)
        now = self._now_fn()
        if full_key in self._cache:
            # Delete so we move to the end of the cache
            del self._cache[full_key]
        self._cache[full_key] = (now, value)
        self._prune(now)



import dataclasses
import typing
import weave
from .. import panel

ModeOption = typing.Literal["plaintext", "markdown", "diff"]


@weave.type()
class PanelStringConfig:
    mode: typing.Optional[ModeOption] = None


@weave.type()
class PanelString(panel.Panel):
    id = "string"
    config: typing.Optional[PanelStringConfig] = dataclasses.field(
        default_factory=lambda: None
    )

    def __init__(self, input_node, vars=None, config=None, **options):
        super().__init__(input_node=input_node, vars=vars)
        self.config = config
        if self.config is None:
            self.config = PanelStringConfig()
        if "mode" in options:
            self.config.mode = options["mode"]



"""
This file contains the Weave Types and Mappers needed to handle tagged values.

Here we will briefly describe how TaggedValues work. Firstly, note that we need
to be able to associate any object in the python runtime with a set of tags. Due
to the way Python works, the best way to implement this is to use a memory
mapping from the python id of an object to a dictionary of tags. This is
implemented in tag_store.py. Unlike other Weave Types, we cannot determine if an
object is tagged simply by looking at its python type, so the TaggedValueType
has custom logic to determine if an object is a TaggedValue as well as special
assignability rules. Furthermore when serializing the TaggedValue to disk, we
need to serialize the tags as well - this is handled between the type class and
the mappers below.

The net result of this system is that op resolvers can operate entirely agnostic
to tags - there is no mutation of the underlying python object whatsoever.
However, the op resolver can still access the tags by using the `find_tag`
function in tag_store.py. This is used by tag getters. Ops can add tags by using
the `add_tags` function in tag_store.py. This is used by tag setters (like
list-tagCheckpoint). Ops that are core language ops will need to operate on tags
directly (for example, list-concat). And ops that are weavifiable will get all
internal tag handling for free (although, this is not yet implemented).
Currently, the degenerate case is that a custom op either copies the object
(creating a new memory address) or is not weavifiable, in which case we will simply
drop tags.
"""

import dataclasses
import json
import typing
import functools

from ... import box
from ... import weave_types as types
from ... import mappers_python
from ... import errors
from ... import mappers

from . import tag_store

if typing.TYPE_CHECKING:
    from ... import artifact_base
    from ... import artifact_fs


def flatten_tag_type_to_typed_dict(tag_type: types.Type) -> types.TypedDict:
    """
    In Weave0, we have a more loose definition of what a TaggedValue is. Specifically,
    a TaggedValue type in Weave0 must conform to the following:
        * Tag: Anything, but in practice is either (or a union of):
                * A TypedDict
                * A TaggedValue (where the value side of the chain are TypedDicts)
        * Value: Must not be a TaggedValue

    This means your data shape looks like a chain of TaggedValues. Consider the following:

    b_0 = type_0
    l_1 = Tag({t_1: tv_1}, b_0) =
    Tagged<
        Dict<
            t_1: tv_1
        >,
        b_0
    >
    l_2 = Tag({t_2: tv_2}, l_1) =
    Tagged<
        Tagged<
            Dict<
                t_1: tv_1
            >,
            Dict<
                t_2: tv_2
            >
        >,
        b_0
    >
    l_3 = Tag({t_3: tv_3}, l_2) =
    Tagged<
        Tagged<
            Tagged<
                Dict<
                    t_1: tv_1
                >,
                Dict<
                    t_2: tv_2
                >
            >,
            Dict<
                t_3: tv_3
            >
        >,
        b_0
    >

    When deciding if a type has a specific tag, we traverse the tag chain. essentially,
    perform the following logic:

    def find_tag(obj):
        if isinstance(obj, TaggedValue):
            return _find_tag_inner(obj.tag)
        else:
            return None

    def _find_tag_inner(obj):
        if isinstance(type.val, TypedDict) and tag_name in type.val:
            return type.val[tag_name]
        elif isinstance(type.tag, TypedDict) and tag_name in type.tag:
            return type.tag[tag_name]
        elif isinstance(type.tag, TaggedValue):
            return _find_tag_inner(type.tag, tag_name)
        else:
            return None

    However, in Weave1, we have a more strict definition of what a TaggedValue is. Specifically:
        * Tag: Must be a TypedDict (values can be TaggedValues)
        * Value: Must not be a TaggedValue

    This means that we need to flatten the tag chain into a single TypedDict.The primary
    disadvantage of this approach to Weave0 is that if we have a tag-key collision then
    only the most recent tag is available. So, the above example becomes:
    Tagged<
        Dict<
            t_1: tv_1,
            t_2: tv_2,
            t_3: tv_3
        >,
        b_0
    >

    This is not strictly necessary - i think if we want we can convert to Weave0's
    definition of TaggedValue. However, this is much nicer to work with and understand.


    This function is used to perform the above described flattening of the tag-side of the
    TaggedValue chain.
    """

    # First, if the tag is already a TypedDict, then we are good
    if isinstance(tag_type, types.TypedDict):
        return tag_type

    # Next, if the tag_type is a union, then we want to merge the results.
    # This is subtly different than Weave0, but ensures that the result
    # is always a TypedDict.
    elif isinstance(tag_type, types.UnionType):
        base_type = types.TypedDict({})
        for member in tag_type.members:
            if not isinstance(member, types.NoneType):
                base_type = types.merge_types(base_type, flatten_tag_type_to_typed_dict(member))  # type: ignore
        return base_type

    # Finally, if the tag_type is a TaggedValue, then we want to merge the tag
    # and value props together.
    elif isinstance(tag_type, TaggedValueType):
        assert types.TypedDict({}).assign_type(tag_type.tag), (
            "tag_type.tag must be assignable to TypedDict, found %s" % tag_type.tag
        )
        assert types.TypedDict({}).assign_type(tag_type.value), (
            "tag_type.value must be assignable to TypedDict, found %s" % tag_type.value
        )
        return types.TypedDict(
            {**tag_type.tag.property_types, **tag_type.value.property_types}  # type: ignore
        )
    else:
        raise errors.WeaveTypeError(
            f"tag_type must be TypedDict, UnionType, or TaggedValueType, found {tag_type}"
        )


# A custom Weave Type used to represent tagged values.
@dataclasses.dataclass(frozen=True)
class TaggedValueType(types.Type):
    name = "tagged"
    tag: types.TypedDict = dataclasses.field(
        default_factory=lambda: types.TypedDict({})
    )
    value: types.Type = dataclasses.field(default_factory=lambda: types.Any())

    _assignment_form_cached = None

    # We use this technique to apply post-processing to the inputs, but also works
    # around the frozen dataclass issue.
    def __post_init__(self) -> None:
        if isinstance(self.value, TaggedValueType):
            self.__dict__["tag"] = types.TypedDict(
                {
                    **self.tag.property_types,
                    **self.value.tag.property_types,
                }
            )
            if isinstance(self.value.value, TaggedValueType):
                raise errors.WeaveTypeError(
                    f"TaggedValueType value types cannot be TaggedValueType, found {self.value.value}"
                )
            self.__dict__["value"] = self.value.value

        # This post processing step is to handle the case where the value is a
        # union of None and a TaggedValueType. This is a common pattern in
        # Weave. In this step, we hoist the tags in the inner TaggedValueType up
        # to the outer TaggedValueType and make them optional, leaving the union
        # to be a union of None and the TaggedValueType's inner value type.
        # While this seems highly specific, the motivation is to solve a bug in
        # Arrow.
        #
        # Firstly, let's justify why this is allowed in the first place. Take
        # the following type: Tagged<{TO},Maybe<Tagged<TI, TV>>. The
        # transformation ends up being: Tagged<{TO, Maybe<TI>},Maybe<TV>>. Now,
        # from a Weave0 compatibility perspective, this is fine because anything
        # the original type is assignable to, the new type is also assignable
        # to. Therefore we are making a slightly broader type here.
        #
        # In fact, I think we can generalize this even farther to always merge
        # out inner unions of TaggedValues.
        #
        # From an op perspective, this fits our model since any tag get
        # operation on the original type will result in an optional value if the
        # tag is on the inner TaggedValue. So, this is perfectly compatible.
        #
        # Now, why do this? The main motivation is AWL. When adding tags to AWL,
        # we use a helper function `direct_add_arrow_tags` which will merge tags
        # in an attempt to conform to the post processing rule above. However,
        # there is one edge case which is incorrect: when the inner value is
        # this optional tagged value case. The reason is because the arrow's
        # `nullable` property on fields is not the same as the Weave union None
        # type. So we have no way of knowing if there is an intermediate "None"
        # and therefore will always merge the tags. This creates a bug when
        # doing type traversal because the Weave type does not match the
        # underlying data shape anymore. This post processing step ensures the
        # types are always aligned.
        #
        # As a bonus, this actually makes the types more accurate with respect
        # to how they are stored in the op_store as well. In the op_store, we
        # merge types in a similar manner, but since we don't need to do any
        # column mapping on the op_store data we never ran into this problem.
        if isinstance(self.value, types.UnionType):
            if len(self.value.members) == 2:
                maybe_none_type = None
                maybe_tagged_value_type = None
                for member in self.value.members:
                    if isinstance(member, types.NoneType):
                        maybe_none_type = member
                    elif isinstance(member, TaggedValueType):
                        maybe_tagged_value_type = member
                if maybe_none_type and maybe_tagged_value_type:
                    new_tag = types.TypedDict(
                        {
                            **self.tag.property_types,
                            **{
                                k: types.optional(v)
                                for k, v in maybe_tagged_value_type.tag.property_types.items()
                            },
                        }
                    )
                    new_value = maybe_tagged_value_type.value

                    self.__dict__["tag"] = new_tag
                    self.__dict__["value"] = new_value

    def simple_str(self) -> str:
        return str(self.value)

    @functools.cached_property
    def _assignment_form(self) -> types.Type:
        if isinstance(self.value, types.UnionType):
            return types.union(
                *[TaggedValueType(self.tag, mem) for mem in self.value.members]
            )
        return self

    def _is_assignable_to(self, other_type: types.Type) -> typing.Optional[bool]:
        if other_type.__class__ != TaggedValueType:
            return other_type.assign_type(self.value)
        return None

    def __getattr__(self, attr: str) -> typing.Any:
        # When getting a property from a tagged value, we want the request to
        # flow through to the inner type. However, if we are getting a type
        # property, we want to wrap the type correctly on the way out. There are
        # two special cases: 1 for property_types and 1 for members. (typeddict
        # and union).
        res = getattr(self.value, attr)
        if isinstance(res, types.Type):
            res = TaggedValueType(self.tag, res)
        elif attr == "property_types" and isinstance(res, dict):
            res = {k: TaggedValueType(self.tag, v) for k, v in res.items()}
        elif attr == "members" and isinstance(res, list):
            res = [TaggedValueType(self.tag, v) for v in res]
        return res

    @classmethod
    def is_instance(cls, obj: typing.Any) -> bool:
        return box.is_boxed(obj) and tag_store.is_tagged(obj)

    @classmethod
    def type_of_instance(cls, obj: typing.Any) -> "TaggedValueType":
        obj = box.box(obj)
        tags = tag_store.get_tags(obj)
        with tag_store.with_visited_obj(obj):
            tag_type = types.TypeRegistry.type_of(tags)
            assert isinstance(tag_type, types.TypedDict), (
                "Tags must be a dictionary, found %s" % tag_type
            )
            value_type = types.TypeRegistry.type_of(obj)
        res = cls(
            tag_type,
            value_type,
        )
        return res

    @classmethod
    def from_dict(cls, d: dict) -> "TaggedValueType":
        tag_type = flatten_tag_type_to_typed_dict(
            types.TypeRegistry.type_from_dict(d["tag"])
        )
        value_type = types.TypeRegistry.type_from_dict(d["value"])
        res = cls(
            tag_type,
            value_type,
        )
        return res

    def _to_dict(self) -> dict:
        return {"tag": self.tag.to_dict(), "value": self.value.to_dict()}

    def save_instance(
        self, obj: types.Any, artifact: "artifact_fs.FilesystemArtifact", name: str
    ) -> None:
        serializer = mappers_python.map_to_python(self, artifact)

        result = serializer.apply(obj)
        with artifact.new_file(f"{name}.object.json") as f:
            json.dump(result, f, allow_nan=False)

    def load_instance(
        self,
        artifact: "artifact_fs.FilesystemArtifact",
        name: str,
        extra: typing.Optional[list[str]] = None,
    ) -> typing.Any:
        with artifact.open(f"{name}.object.json") as f:
            result = json.load(f)
        mapper = mappers_python.map_from_python(self, artifact)
        return mapper.apply(result)

    def __repr__(self) -> str:
        return (
            f"TaggedValueType({{{list(self.tag.property_types.keys())}}}, {self.value})"
        )


class TaggedValueMapper(mappers.Mapper):
    def __init__(
        self,
        type_: TaggedValueType,
        mapper: typing.Callable,  # TODO: Make this more specific
        artifact: "artifact_base.Artifact",
        path: list[str] = [],
    ):
        self.type = type_
        self._artifact = artifact
        self._tag_serializer = mapper(type_.tag, artifact, path=path + ["_tag"])
        self._value_serializer = mapper(type_.value, artifact, path=path + ["_value"])


class TaggedValueToPy(TaggedValueMapper):
    def apply(self, obj: typing.Any) -> dict:
        result = {}
        obj_tags = tag_store.get_tags(obj)
        expected_tag_keys = set(
            key
            for key, tag_type in self.type.tag.property_types.items()
            if not types.is_optional(tag_type)
        )
        if len(expected_tag_keys - set(obj_tags.keys())) > 0:
            raise errors.WeaveTypeError(
                f"Expected tags {self.type.tag.property_types.keys()}, found {obj_tags.keys()}"
            )
        result["_tag"] = self._tag_serializer.apply(obj_tags)
        result["_value"] = self._value_serializer.apply(obj)
        return result


class TaggedValueFromPy(TaggedValueMapper):
    def apply(self, obj: dict) -> typing.Any:
        value = self._value_serializer.apply(obj["_value"])
        tags = self._tag_serializer.apply(obj["_tag"])
        value = box.box(value)
        tag_store.add_tags(value, tags)
        return value



# Cache policy for when cache mode is minimal.
# We don't declare these directly on op defs for now. I want op definitions
# to be more declarative than that. Ie they should be cached if they are
# known to be expensive, either because of an API call that costs $ or
# requiring a lot of compute/memory.

# These operate on "simple_name" (non uri name), which is called "full_name"
# by graph.py


CACHE_AND_PARALLEL_OP_NAMES = [
    "op-expensive_op",
    "Chain-run",
    "BaseChatModel-predict",
    "op-stable_diffusion",
    "op-img2prompt",
]

CACHE_NON_PURE_OP_NAMES = [
    "gqlroot-wbgqlquery",
]

CACHE_OP_NAMES = (
    [
        "file-readcsv",
        "op-run_chain",
        "table-2DProjection",
        "table-projection2D",
        "ArrowWeaveList-2DProjection",
        "ArrowWeaveList-projection2D",
        "op-faiss_from_documents",
        "FAISS-document_embeddings",
        "HuggingFacePackage-model",
        "HFModelTextClassification-pipeline",
        "HFModelTextClassification-call",
        # These are parallelized by derive_op only, in a custom way
        # Should move them to CACHE_AND_PARALLEL_OP_NAMES once that
        # is fixed.
        "file-partitionedTable",
        "file-table",
        "file-joinedTable",
        "op-umap_project",
        "op-openai_embed",
        "op-hdbscan_cluster",
        "wb_trace_tree-convertToSpans",
    ]
    + CACHE_AND_PARALLEL_OP_NAMES
    + CACHE_NON_PURE_OP_NAMES
)

ARROW_FS_OPS = [
    "run-history3",
    "run-history3_with_columns",
    "table-rows",
    "partitionedtable-rows",
    "joinedtable-rows",
    "mapped_table-rows",
    "mapped_partitionedtable-rows",
    "mapped_joinedtable-rows",
]


# history ops are parallelized by derive_op only, in a custom way
# PARALLEL_OP_NAMES = ["run-history", "run-history2"] + CACHE_AND_PARALLEL_OP_NAMES
PARALLEL_OP_NAMES = CACHE_AND_PARALLEL_OP_NAMES


def should_run_in_parallel(op_name: str) -> bool:
    # Uncomment to enable parallelism for custom ops (weaveflow)
    # if "://" in op_name:
    #     # Always parallelize custom ops for now (ops that are not built-ins)
    #     return True
    if op_name.startswith("mapped_"):
        op_name = op_name[len("mapped_") :]
    return op_name in PARALLEL_OP_NAMES


def should_cache(op_name: str) -> bool:
    if op_name.startswith("mapped_"):
        return False
    return op_name in CACHE_OP_NAMES


def should_table_cache(op_name: str) -> bool:
    return False
